# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from absl.testing import absltest
from absl.testing import parameterized
import tensorflow.compat.v1 as tf

from open_spiel.python import rl_environment
from open_spiel.python.algorithms import policy_gradient
from open_spiel.python.algorithms.losses import rl_losses
import pyspiel


class PolicyGradientTest(parameterized.TestCase, tf.test.TestCase):

  @parameterized.parameters(
      itertools.product(("rpg", "qpg", "rm", "a2c"),
                        ("kuhn_poker", "leduc_poker")))
  def test_run_game(self, loss_str, game_name):
    env = rl_environment.Environment(game_name)
    info_state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    with self.session() as sess:
      agents = [
          policy_gradient.PolicyGradient(  # pylint: disable=g-complex-comprehension
              sess,
              player_id=player_id,
              info_state_size=info_state_size,
              num_actions=num_actions,
              loss_str=loss_str,
              hidden_layers_sizes=[8, 8],
              batch_size=16,
              entropy_cost=0.001,
              critic_learning_rate=0.01,
              pi_learning_rate=0.01,
              num_critic_before_pi=4) for player_id in [0, 1]
      ]
      sess.run(tf.global_variables_initializer())

      for _ in range(2):
        time_step = env.reset()
        while not time_step.last():
          current_player = time_step.observations["current_player"]
          current_agent = agents[current_player]
          agent_output = current_agent.step(time_step)
          time_step = env.step([agent_output.action])

        for agent in agents:
          agent.step(time_step)

  @absltest.skip("Causing a segmentation fault on wheel tests")
  def test_run_hanabi(self):
    # Hanabi is an optional game, so check we have it before running the test.
    game = "hanabi"
    if game not in pyspiel.registered_names():
      return

    num_players = 3
    env_configs = {
        "players": num_players,
        "max_life_tokens": 1,
        "colors": 2,
        "ranks": 3,
        "hand_size": 2,
        "max_information_tokens": 3,
        "discount": 0.
    }
    env = rl_environment.Environment(game, **env_configs)
    info_state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    with self.session() as sess:
      agents = [
          policy_gradient.PolicyGradient(  # pylint: disable=g-complex-comprehension
              sess,
              player_id=player_id,
              info_state_size=info_state_size,
              num_actions=num_actions,
              hidden_layers_sizes=[8, 8],
              batch_size=16,
              entropy_cost=0.001,
              critic_learning_rate=0.01,
              pi_learning_rate=0.01,
              num_critic_before_pi=4) for player_id in range(num_players)
      ]
      sess.run(tf.global_variables_initializer())
      time_step = env.reset()
      while not time_step.last():
        current_player = time_step.observations["current_player"]
        agent_output = [agent.step(time_step) for agent in agents]
        time_step = env.step([agent_output[current_player].action])

      for agent in agents:
        agent.step(time_step)

  def test_loss_modes(self):
    loss_dict = {
        "qpg": rl_losses.BatchQPGLoss,
        "rpg": rl_losses.BatchRPGLoss,
        "rm": rl_losses.BatchRMLoss,
        "a2c": rl_losses.BatchA2CLoss,
    }
    with self.session() as sess:
      for loss_str, loss_class in loss_dict.items():
        agent_by_str = policy_gradient.PolicyGradient(
            sess,
            player_id=0,
            info_state_size=32,
            num_actions=2,
            loss_str=loss_str,
            loss_class=None)
        agent_by_class = policy_gradient.PolicyGradient(
            sess,
            player_id=0,
            info_state_size=32,
            num_actions=2,
            loss_str=None,
            loss_class=loss_class)

        self.assertEqual(agent_by_str._pi_loss.shape,
                         agent_by_class._pi_loss.shape)
        self.assertEqual(agent_by_str._pi_loss.dtype,
                         agent_by_class._pi_loss.dtype)
        self.assertEqual(agent_by_str._pi_loss.op.type,
                         agent_by_class._pi_loss.op.type)


if __name__ == "__main__":
  tf.test.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stochastic Gradient Descent (Adam) Approx. Nash Solver w/ Annealing."""

from absl import logging  # pylint:disable=unused-import

import jax
import jax.numpy as jnp

import numpy as np

import optax

from scipy import special

from open_spiel.python.algorithms.adidas_utils.helpers import simplex
from open_spiel.python.algorithms.adidas_utils.helpers.nonsymmetric import exploitability as exp


class Solver(object):
  """Adam Solver with temperature annealing."""

  def __init__(self, temperature=1., proj_grad=True, lrs=(1e-2, 1e-1),
               exp_thresh=-1., rnd_init=False, seed=None, **kwargs):
    """Ctor."""
    del kwargs
    if temperature < 0.:
      raise ValueError("temperature must be non-negative")
    self.num_players = None
    self.temperature = temperature
    self.proj_grad = proj_grad
    self.rnd_init = rnd_init
    self.lrs = lrs
    self.num_estimates = 2
    self.exp_thresh = exp_thresh
    self.has_aux = True
    self.aux_errors = []

    self.update = self.descent_step

    self.opt = optax.adam(learning_rate=lrs[0])
    self.opt_state = self.opt.init(jnp.zeros(1))

    self.seed = seed
    self.random = np.random.RandomState(seed)

  def init_vars(self, num_strats, num_players):
    """Initialize solver parameters."""
    self.num_players = num_players
    if len(num_strats) != num_players:
      raise ValueError("Must specify num strategies for each player")

    init_dist = []
    for num_strats_i in num_strats:
      if self.rnd_init:
        init_dist_i = self.random.rand(num_strats_i)
      else:
        init_dist_i = np.ones(num_strats_i)
      init_dist_i /= init_dist_i.sum()
      init_dist.append(init_dist_i)

    init_params = [
        jnp.array(dist_to_logits(init_dist_i)) for init_dist_i in init_dist
    ]

    self.opt_state = self.opt.init(init_params)

    init_y = [np.zeros_like(dist_i) for dist_i in init_dist]
    init_anneal_steps = 0

    return (init_dist, init_y, init_anneal_steps)

  def record_aux_errors(self, grads):
    """Record errors for the auxiliary variables."""
    grad_y = grads[1]
    # call ravel in case use y to track entire payoff matrices in future
    grad_y_flat = np.concatenate([np.ravel(g) for g in grad_y])
    self.aux_errors.append([np.linalg.norm(grad_y_flat)])

  def compute_gradients(self, params, payoff_matrices):
    """Compute and return gradients (and exploitabilities) for all parameters.

    Args:
      params: tuple of params (dist, y, anneal_steps), see gradients
      payoff_matrices: 2 dictionaries with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    Returns:
      tuple of gradients (grad_dist, grad_y, grad_anneal_steps), see gradients
      unregularized exploitability (stochastic estimate)
      shannon entropy regularized exploitability (stochastic estimate)
    """
    return self.gradients(*params, payoff_matrices, self.num_players,
                          self.temperature, self.proj_grad)

  def exploitability(self, params, payoff_matrices):
    """Compute and return shannon entropy regularized exploitability.

    Args:
      params: tuple of params (dist, y), see qre.gradients
      payoff_matrices: 2 dictionaries with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    Returns:
      float, exploitability of current dist
    """
    return exp.qre_exploitability(params, payoff_matrices, self.temperature)

  def gradients(self, dist: np.ndarray, y: np.ndarray, anneal_steps: int,
                payoff_matrices, num_players,
                temperature=0., proj_grad=True
                ) -> tuple[tuple[list[np.ndarray], list[np.ndarray], int],
                           float,
                           float]:
    """Computes exploitablity gradient and aux variable gradients.

    Args:
      dist: list of 1-d np.arrays, current estimate of nash distribution
      y: list 1-d np.arrays (same shape as dist), current est. of payoff
        gradient
      anneal_steps: int, elapsed num steps since last anneal
      payoff_matrices: 2 dictionaries with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
      num_players: int, number of players, in case payoff_matrices is
        abbreviated
      temperature: non-negative float, default 0.
      proj_grad: bool, if True, projects dist gradient onto simplex
    Returns:
      gradient of exploitability w.r.t. (dist, y, anneal_steps) as tuple
      unregularized exploitability (stochastic estimate)
      shannon entropy regularized exploitability (stochastic estimate)
    """

    grad_dist = loss_gradients(dist, payoff_matrices, num_players, temperature,
                               proj_grad)[0][0]

    grad_y = []
    unreg_exp = []
    reg_exp = []
    for i in range(num_players):

      nabla_i = np.zeros_like(dist[i])
      for j in range(num_players):
        if j == i:
          continue
        if i < j:
          hess_i_ij = 0.5 * payoff_matrices[0][(i, j)][0]
          hess_i_ij += 0.5 * payoff_matrices[1][(i, j)][0]
        else:
          hess_i_ij = 0.5 * payoff_matrices[0][(j, i)][1].T
          hess_i_ij += 0.5 * payoff_matrices[1][(j, i)][1].T

        nabla_ij = hess_i_ij.dot(dist[j])
        nabla_i += nabla_ij / float(num_players - 1)

      grad_y.append(y[i] - nabla_i)

      if temperature >= 1e-3:
        br_i = special.softmax(y[i] / temperature)
      else:
        power = np.inf
        s_i = np.linalg.norm(y[i], ord=power)
        br_i = np.zeros_like(dist[i])
        maxima_i = (y[i] == s_i)
        br_i[maxima_i] = 1. / maxima_i.sum()

      unreg_exp.append(np.max(y[i]) - y[i].dot(dist[i]))

      entr_br_i = temperature * special.entr(br_i).sum()
      entr_dist_i = temperature * special.entr(dist[i]).sum()

      reg_exp.append(y[i].dot(br_i - dist[i]) + entr_br_i - entr_dist_i)

    unreg_exp_mean = np.mean(unreg_exp)
    reg_exp_mean = np.mean(reg_exp)

    _, lr_y = self.lrs
    if (reg_exp_mean < self.exp_thresh) and (anneal_steps >= 1 / lr_y):
      self.temperature = np.clip(temperature / 2., 0., np.inf)
      grad_anneal_steps = -anneal_steps
    else:
      grad_anneal_steps = 1

    return (grad_dist, grad_y, grad_anneal_steps), unreg_exp_mean, reg_exp_mean

  def descent_step(self, params, grads, t, eps=0.):
    """Gradient descent on exploitability wrt logits.

    Args:
      params: tuple of variables to be updated (dist, y, anneal_steps)
      grads: tuple of variable gradients (grad_dist, grad_y, grad_anneal_steps)
      t: int, solver iteration
      eps: float > 0, force all probabilities >= eps / dim(dist) (unused)
    Returns:
      new_params: tuple of update params (new_dist, new_y, new_anneal_steps)
    """
    del eps

    dist = params[0]
    grads_dist = grads[0]

    dist_jnp = [jnp.array(dist_i) for dist_i in dist]
    grads_dist_jnp = [jnp.array(grad_i) for grad_i in grads_dist]

    # map dist to logits and grads to grad_logits using jacobian
    logits = [dist_to_logits(dist_i) for dist_i in params[0]]
    grads_logits = [
        jax.jvp(dist_to_logits, [dist_i], [grads_i])[1]
        for dist_i, grads_i in zip(dist_jnp, grads_dist_jnp)
    ]

    opt_updates, self.opt_state = self.opt.update(grads_logits,
                                                  self.opt_state,
                                                  logits)

    new_logits = optax.apply_updates(logits, opt_updates)

    new_dist = [logits_to_dist(logits) for logits in new_logits]
    new_dist = [np.array(dist_i) for dist_i in new_dist]

    lr_y = self.lrs[1]
    lr_y = np.clip(1 / float(t + 1), lr_y, np.inf)
    new_y = []
    for y_i, y_grad_i in zip(params[1], grads[1]):
      new_y_i = y_i - lr_y * y_grad_i
      new_y_i = np.clip(new_y_i, 0., np.inf)
      new_y.append(new_y_i)

    new_anneal_steps = params[2] + grads[2]

    return (new_dist, new_y, new_anneal_steps)


def logits_to_dist(logits):
  logits_ext = jnp.append(logits, 0.)
  payoff = jax.nn.softmax(logits_ext)
  return payoff


def dist_to_logits(dist, eps=1e-8):
  # dist[-1] = exp(logits[-1]) / Z = exp(0) / Z
  z = 1 / jnp.clip(dist[-1], eps, 1.)
  logits = jnp.log(jnp.clip(dist[:-1] * z, eps, np.inf))
  return logits


def loss_gradients(dist, payoff_matrices, num_players, temperature=0.,
                   proj_grad=True):
  """Computes exploitablity gradient.

  Args:
    dist: list of 1-d np.arrays, current estimate of nash distribution
    payoff_matrices: 2 dictionaries with keys as tuples of agents (i, j) and
        values of (2 x A x A) np.arrays, payoffs for each joint action. keys
        are sorted and arrays should be indexed in the same order
    num_players: int, number of players, in case payoff_matrices is abbreviated
    temperature: non-negative float, default 0.
    proj_grad: bool, if True, projects dist gradient onto simplex
  Returns:
    gradient of exploitability w.r.t. (dist) as tuple
    unregularized exploitability (stochastic estimate)
    shannon regularized exploitability (stochastic estimate)
  """
  # first compute projected gradients (for every player, for each sample a & b)
  # if consulting paper https://arxiv.org/abs/2310.06689, code assumes eta_k = 1
  tau = temperature

  pgs = []
  for i in range(num_players):

    pg_i_a = np.zeros_like(dist[i])
    pg_i_b = np.zeros_like(dist[i])

    for j in range(num_players):
      if j == i:
        continue
      if i < j:
        hess_i_ij_a = payoff_matrices[0][(i, j)][0]
        hess_i_ij_b = payoff_matrices[1][(i, j)][0]
      else:
        hess_i_ij_a = payoff_matrices[0][(j, i)][1].T
        hess_i_ij_b = payoff_matrices[1][(j, i)][1].T

      pg_i_a_est = simplex.project_grad(hess_i_ij_a.dot(dist[j]))
      pg_i_b_est = simplex.project_grad(hess_i_ij_b.dot(dist[j]))

      pg_i_a += pg_i_a_est / float(num_players - 1)
      pg_i_b += pg_i_b_est / float(num_players - 1)

    pgs.append((pg_i_a, pg_i_b))

  # then construct unbiased stochastic gradient
  grad_dist = []
  unreg_exp = []
  reg_exp = []

  for i in range(num_players):

    grad_dist_i = np.zeros_like(dist[i])

    for j in range(num_players):
      pg_j_a = pgs[j][0]
      pg_j_b = pgs[j][1]
      if tau > 0.:
        log_dist_safe = np.clip(np.log(dist[j]), -40, 0)
        entr_grad_proj = simplex.project_grad(-tau * (log_dist_safe + 1))
      else:
        entr_grad_proj = 0.
      pg_j_a_entr = pg_j_a + entr_grad_proj
      pg_j_b_entr = pg_j_b + entr_grad_proj

      if j == i:
        if tau > 0.:
          hess_j_ij_a = -tau * np.diag(1. / dist[j])
        else:
          hess_j_ij_a = np.diag(np.zeros_like(dist[j]))
        unreg_exp_i = np.dot(pg_j_a, pg_j_b)
        reg_exp_i = np.dot(pg_j_a_entr, pg_j_b_entr)
        unreg_exp.append(unreg_exp_i)
        reg_exp.append(reg_exp_i)
      elif i < j:
        hess_j_ij_a = payoff_matrices[0][(i, j)][1]
      else:
        hess_j_ij_a = payoff_matrices[0][(j, i)][0].T

      grad_dist_i += 2. * hess_j_ij_a.dot(pg_j_b_entr)

    if proj_grad:
      grad_dist_i = simplex.project_grad(grad_dist_i)

    grad_dist.append(grad_dist_i)

  return (grad_dist,), np.mean(unreg_exp), np.mean(reg_exp)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Computes the distribution of a policy."""
import collections

from typing import List, Tuple
from open_spiel.python import policy as policy_module
from open_spiel.python.mfg import tabular_distribution
from open_spiel.python.mfg.tabular_distribution import DistributionDict
import pyspiel


def type_from_states(states):
  """Get node type of a list of states and assert they are the same."""
  types = [state.get_type() for state in states]
  assert len(set(types)) == 1, f"types: {types}"
  return types[0]


def _check_distribution_sum(distribution: DistributionDict, expected_sum: int):
  """Sanity check that the distribution sums to a given value."""
  sum_state_probabilities = sum(distribution.values())
  assert abs(sum_state_probabilities - expected_sum) < 1e-4, (
      "Sum of probabilities of all possible states should be the number of "
      f"population, it is {sum_state_probabilities}.")


class DistributionPolicy(tabular_distribution.TabularDistribution):
  """Computes the distribution of a specified strategy."""

  def __init__(self,
               game: pyspiel.Game,
               policy: policy_module.Policy,
               root_state: pyspiel.State = None):
    """Initializes the distribution calculation.

    Args:
      game: The game to analyze.
      policy: The policy we compute the distribution of.
      root_state: The state of the game at which to start analysis. If `None`,
        the game root states are used.
    """
    super().__init__(game)
    self._policy = policy
    if root_state is None:
      self._root_states = game.new_initial_states()
    else:
      self._root_states = [root_state]
    self.evaluate()

  def evaluate(self):
    """Evaluate the distribution over states of self._policy."""
    # List of all game states that have a non-zero probability at the current
    # timestep and player ID.
    current_states = self._root_states.copy()
    # Distribution at the current timestep. Maps state strings to
    # floats. For each group of states for a given population, these
    # floats represent a probability distribution.
    current_distribution = {
        self.state_to_str(state): 1 for state in current_states
    }
    # List of all distributions computed so far.
    all_distributions = [current_distribution]

    while type_from_states(current_states) != pyspiel.StateType.TERMINAL:
      new_states, new_distribution = self._one_forward_step(
          current_states, current_distribution, self._policy)
      _check_distribution_sum(new_distribution, self.game.num_players())
      current_distribution = new_distribution
      current_states = new_states
      all_distributions.append(new_distribution)

    # Merge all per-timestep distributions into `self.distribution`.
    for dist in all_distributions:
      for state_str, prob in dist.items():
        if state_str in self.distribution:
          raise ValueError(
              f"{state_str} has already been seen in distribution.")
        self.distribution[state_str] = prob

  def _forward_actions(
      self, current_states: List[pyspiel.State], distribution: DistributionDict,
      actions_and_probs_fn) -> Tuple[List[pyspiel.State], DistributionDict]:
    """Applies one action to each current state.

    Args:
      current_states: The states to apply actions on.
      distribution: Current distribution.
      actions_and_probs_fn: Function that maps one state to the corresponding
        list of (action, proba). For decision nodes, this should be the policy,
        and for chance nodes, this should be chance outcomes.

    Returns:
      A pair:
        - new_states: List of new states after applying one action on
          each input state.
        - new_distribution: Probabilities for each of these states.
    """
    new_states = []
    new_distribution = collections.defaultdict(float)
    for state in current_states:
      state_str = self.state_to_str(state)
      for action, prob in actions_and_probs_fn(state):
        new_state = state.child(action)
        new_state_str = self.state_to_str(new_state)
        if new_state_str not in new_distribution:
          new_states.append(new_state)
        new_distribution[new_state_str] += prob * distribution[state_str]
    return new_states, new_distribution

  def _one_forward_step(self, current_states: List[pyspiel.State],
                        distribution: DistributionDict,
                        policy: policy_module.Policy):
    """Performs one step of the forward equation.

    Namely, this takes as input a list of current state, the current
    distribution, and performs one step of the forward equation, using
    actions coming from the policy or from the chance node
    probabilities, or propagating the distribution to the MFG nodes.

    Args:
      current_states: The states to perform the forward step on. All states are
        assumed to be of the same type.
      distribution: Current distribution.
      policy: Policy that will be used if states

    Returns:
      A pair:
        - new_states: List of new states after applying one step of the
          forward equation (either performing one action or doing one
          distribution update).
        - new_distribution: Probabilities for each of these states.
    """
    state_types = type_from_states(current_states)
    if state_types == pyspiel.StateType.CHANCE:
      return self._forward_actions(current_states, distribution,
                                   lambda state: state.chance_outcomes())

    if state_types == pyspiel.StateType.MEAN_FIELD:
      new_states = []
      new_distribution = {}
      for state in current_states:
        dist = [
            # We need to default to 0, since the support requested by
            # the state in `state.distribution_support()` might have
            # states that we might not have reached yet. A probability
            # of 0. should be given for them.
            distribution.get(str_state, 0.)
            for str_state in state.distribution_support()
        ]
        new_state = state.clone()
        new_state.update_distribution(dist)
        new_state_str = self.state_to_str(new_state)
        if new_state_str not in new_distribution:
          new_states.append(new_state)
          new_distribution[new_state_str] = 0.0
        new_distribution[new_state_str] += distribution.get(
            self.state_to_str(state), 0)
      return new_states, new_distribution

    if state_types == pyspiel.StateType.DECISION:
      return self._forward_actions(
          current_states, distribution,
          lambda state: policy.action_probabilities(state).items())

    raise ValueError(
        f"Unpexpected state_stypes: {state_types}, states: {current_states}")



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for computing gradient information: run games and record payoffs.
"""

import itertools

from absl import logging  # pylint:disable=unused-import

import numpy as np


def construct_game_queries(base_profile, num_checkpts):
  """Constructs a list of checkpoint selection tuples to query value function.

  Each query tuple (key, query) where key = (pi, pj) and query is
  (p1's selected checkpt, ..., p7's selected checkpt) fixes the players in the
  game of diplomacy to be played. It may be necessary to play several games with
  the same players to form an accurate estimate of the value or payoff for each
  player as checkpts contain stochastic policies.

  Args:
    base_profile: list of selected checkpts for each player, i.e.,
      a sample from the player strategy profile ([x_i ~ p(x_i)])
    num_checkpts: list of ints, number of strats (or ckpts) per player
  Returns:
    Set of query tuples containing a selected checkpoint index for each player.
  """
  new_queries = set([])

  num_players = len(base_profile)
  for pi, pj in itertools.combinations(range(num_players), 2):
    new_profile = list(base_profile)
    for ai in range(num_checkpts[pi]):
      new_profile[pi] = ai
      for aj in range(num_checkpts[pj]):
        new_profile[pj] = aj
        query = tuple(new_profile)
        pair = (pi, pj)
        new_queries.update([(pair, query)])

  return new_queries


def construct_game_queries_for_exp(base_profile, num_checkpts):
  """Constructs a list of checkpoint selection tuples to query value function.

  Each query tuple (key, query) where key = (pi,) and query is
  (p1's selected checkpt, ..., p7's selected checkpt) fixes the players in the
  game of diplomacy to be played. It may be necessary to play several games with
  the same players to form an accurate estimate of the value or payoff for each
  player as checkpts contain stochastic policies.

  Args:
    base_profile: list of selected checkpts for each player, i.e.,
      a sample from the player strategy profile ([x_i ~ p(x_i)])
    num_checkpts: list of ints, number of strats (or ckpts) per player
  Returns:
    Set of query tuples containing a selected checkpoint index for each player.
  """
  new_queries = set([])

  num_players = len(base_profile)
  for pi in range(num_players):
    new_profile = list(base_profile)
    for ai in range(num_checkpts[pi]):
      new_profile[pi] = ai
      query = tuple(new_profile)
      new_queries.update([(pi, query)])

  return new_queries


def run_games_and_record_payoffs(game_queries, evaluate_game, ckpt_to_policy):
  """Simulate games according to game queries and return results.

  Args:
    game_queries: set of tuples containing indices specifying each players strat
      key_query = (agent_tuple, profile_tuple) format
    evaluate_game: callable function that takes a list of policies as argument
    ckpt_to_policy: list of maps from strat (or checkpoint) to a policy, one
      map for each player
  Returns:
    dictionary: key=key_query, value=np.array of payoffs (1 for each player)
  """
  game_results = {}
  for key_query in game_queries:
    _, query = key_query
    policies = [ckpt_to_policy[pi][ckpt_i] for pi, ckpt_i in enumerate(query)]
    payoffs = evaluate_game(policies)
    game_results.update({key_query: payoffs})
  return game_results


def form_payoff_matrices(game_results, num_checkpts):
  """Packages dictionary of game results into a payoff tensor.

  Args:
    game_results: dictionary of payoffs for each game evaluated, keys are
      (pair, profile) where pair is a tuple of the two agents played against
      each other and profile indicates pure joint action played by all agents
    num_checkpts: list of ints, number of strats (or ckpts) per player
  Returns:
    payoff_matrices: dict of np.arrays (2 x num_checkpts x num_checkpts) with
      payoffs for two players. keys are pairs above with lowest index agent
      first
  """
  num_players = len(num_checkpts)
  payoff_matrices = {}
  for pi, pj in itertools.combinations(range(num_players), 2):
    key = (pi, pj)
    payoff_matrices[key] = np.zeros((2, num_checkpts[pi], num_checkpts[pj]))
  for key_profile, payoffs in game_results.items():
    key, profile = key_profile
    i, j = key
    ai = profile[i]
    aj = profile[j]
    payoff_matrices[key][0, ai, aj] = payoffs[i]
    payoff_matrices[key][1, ai, aj] = payoffs[j]
  return payoff_matrices



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example use of the CFR algorithm on Kuhn Poker."""

import pickle
import sys
from absl import app
from absl import flags

import pyspiel

universal_poker = pyspiel.universal_poker

FLAGS = flags.FLAGS

flags.DEFINE_enum("solver", "cfr", ["cfr", "cfrplus", "cfrbr"], "CFR solver")
_ITERATIONS = flags.DEFINE_integer("iterations", 100, "Number of iterations")

CUSTOM_LIMIT_HOLDEM_ACPC_GAMEDEF = """\
GAMEDEF
limit
numPlayers = 2
numRounds = 1
blind = 2 4
raiseSize = 4 4 8
firstPlayer = 1
maxRaises = 2 2 2
numSuits = 2
numRanks = 5
numHoleCards = 1
numBoardCards = 0 2 1
stack = 20
END GAMEDEF
"""


def main(_):
  game = universal_poker.load_universal_poker_from_acpc_gamedef(
      CUSTOM_LIMIT_HOLDEM_ACPC_GAMEDEF
  )

  solver = None
  if FLAGS.solver == "cfr":
    solver = pyspiel.CFRSolver(game)
  elif FLAGS.solver == "cfrplus":
    solver = pyspiel.CFRPlusSolver(game)
  elif FLAGS.solver == "cfrbr":
    solver = pyspiel.CFRBRSolver(game)
  else:
    print("Unknown solver")
    sys.exit(0)

  for i in range(int(_ITERATIONS.value / 2)):
    solver.evaluate_and_update_policy()
    print("Iteration {} exploitability: {:.6f}".format(
        i, pyspiel.exploitability(game, solver.average_policy())))

  filename = "/tmp/{}_solver.pickle".format(FLAGS.solver)
  print("Persisting the model...")
  with open(filename, "wb") as file:
    pickle.dump(solver, file, pickle.HIGHEST_PROTOCOL)

  print("Loading the model...")
  with open(filename, "rb") as file:
    loaded_solver = pickle.load(file)
  print("Exploitability of the loaded model: {:.6f}".format(
      pyspiel.exploitability(game, loaded_solver.average_policy())))

  for i in range(int(_ITERATIONS.value / 2)):
    loaded_solver.evaluate_and_update_policy()
    tabular_policy = loaded_solver.tabular_average_policy()
    print(f"Tabular policy length: {len(tabular_policy)}")
    print(
        "Iteration {} exploitability: {:.6f}".format(
            int(_ITERATIONS.value / 2) + i,
            pyspiel.exploitability(game, loaded_solver.average_policy()),
        )
    )


if __name__ == "__main__":
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ranked Pairs A.K.A. the Tideman method.

Based on https://en.wikipedia.org/wiki/Ranked_pairs.
"""

from typing import List, Tuple
import numpy as np
from open_spiel.python.voting import base

# TODO(author5): either one of the following: (i) change graph representation to
# adjacency lists for more efficient cycle checking, (ii) use a graph library
# such as networkx to represent the graph and support graph functions.


class RankedPairsRankOutcome(base.RankOutcome):
  """A custom RankOutcome class for Ranked Pairs.

  Provides an extra method to get the graph.
  """

  def __init__(
      self,
      rankings: List[base.AlternativeId],
      scores: List[float],
      graph: np.ndarray,
  ):
    super().__init__(rankings, scores)
    self._graph = graph

  @property
  def graph(self) -> np.ndarray:
    return self._graph


class RankedPairsVoting(base.AbstractVotingMethod):
  """Implements Ranked Pairs / Tideman's method."""

  def __init__(self):
    pass

  def name(self) -> str:
    return "ranked_pairs"

  def _would_create_cycle(
      self,
      alternatives: List[base.AlternativeId],
      graph: np.ndarray,
      from_idx: int,
      to_idx: int,
  ) -> bool:
    """Checks if adding a specific directed edge would result in a cycle.

    Args:
      alternatives: list of alternatives.
      graph: 2D adjacency matrix representing a directed acyclic graph. Row is
        the from node index, column the to node index.
      from_idx: the edge to add (from index).
      to_idx: the edge to add (to index).

    Returns:
      True if adding the specified edge would result in a cycle in the graph.
    """
    # Perform a breadth-first flood fill using a status table.
    # Values in the status table represent:
    #    0  means it does not exist in the flood yet
    #    1  means it needs to be expanded
    #    -1 means it has been expanded (now closed, do not revisit)
    m = len(alternatives)
    status_table = np.zeros(m)
    status_table[to_idx] = 1
    num_expanded = 1
    while num_expanded > 0:
      num_expanded = 0
      for i in np.where(status_table == 1)[0]:
        num_expanded += 1
        for j in np.where(graph[i][:] == 1)[0]:
          if status_table[j] == 0:
            if j == from_idx:
              return True
            status_table[j] = 1
        status_table[i] = -1
    return False

  def _is_source(self, graph: np.ndarray, idx: int):
    """Returns true if this node is a source, false otherwise."""
    num_incoming = np.sum(graph[:, idx])
    num_outgoing = np.sum(graph[idx])
    return num_outgoing > 0 and num_incoming == 0

  def _remove_node(self, graph: np.ndarray, idx: int):
    """Removes a node from the graph."""
    graph[idx, :] = 0
    graph[:, idx] = 0

  def _get_score(
      self, graph: np.ndarray, margin_matrix: np.ndarray, node_idx: int
  ) -> int:
    """Computes the score of an alternative.

    The score is defined as the sum of the margins between the subgraph
    containing all reachable nodes from this node.

    Args:
      graph: 2D adjacency matrix representing a directed acyclic graph. Row is
        the from node index, column the to node index.
      margin_matrix: the margin matrix from the profile
      node_idx: the node index in question.

    Returns:
      the score of the alternative represented by this node index.
    """
    # Flood fill to compute score from a source
    score = 0
    open_list = {node_idx: True}
    closed_list = {}
    while open_list:
      i = list(open_list.keys())[0]
      open_list.pop(i)
      outgoing_edges = np.where(graph[i][:] == 1)[0]
      for j in outgoing_edges:
        score += margin_matrix[i, j]
        if j not in open_list and j not in closed_list:
          open_list[j] = True
      closed_list[i] = True
    return score

  def _get_ranked_pairs(
      self, alternatives: List[base.AlternativeId], margin_matrix: np.ndarray
  ) -> List[Tuple[Tuple[base.AlternativeId, base.AlternativeId], int]]:
    """Returns the positively-valued ranked pairs coupled with their values.
 
    Arguments:
      alternatives: the list of alternatives ids.
      margin_matrix: the margin matrix we use to get the values for each ranked
        pair.

    Returns:
      A list of tuples of the form ((x, y), value) indicating x beating y by
      the specified value.
    """
    ranked_pairs = {}
    rows, cols = np.where(margin_matrix > 0)
    for i, j in zip(rows, cols):
      key_tup = (alternatives[i], alternatives[j])
      ranked_pairs[key_tup] = margin_matrix[i, j]
    return sorted(ranked_pairs.items(), key=lambda item: item[1], reverse=True)

  def run_election(
      self, profile: base.PreferenceProfile
  ) -> RankedPairsRankOutcome:
    assert self.is_valid_profile(profile)
    alternatives = profile.alternatives
    m = len(alternatives)
    alt_idx = profile.alternatives_dict
    margin_matrix = profile.margin_matrix()

    # First, get the ranked pairs annotated with their values (delta(a,b)).
    sorted_pairs = self._get_ranked_pairs(alternatives, margin_matrix)

    # Now, create the graph: add edges that do not create cycles.
    graph = np.zeros(shape=(m, m), dtype=np.int32)
    if sorted_pairs:
      # Create the top-ranked pair. This needs to be in a conditional block,
      # because some profiles can legitimately lead to a graph with no edges (no
      # positively-valued ranked pairs)
      first_pair = sorted_pairs[0][0]
      p0_idx = alt_idx[first_pair[0]]
      p1_idx = alt_idx[first_pair[1]]
      graph[p0_idx, p1_idx] = 1
    for j in range(1, len(sorted_pairs)):
      pair = sorted_pairs[j][0]
      p0_idx = alt_idx[pair[0]]
      p1_idx = alt_idx[pair[1]]
      if not self._would_create_cycle(alternatives, graph, p0_idx, p1_idx):
        graph[p0_idx, p1_idx] = 1
    full_graph = graph.copy()  # Make a copy to return later.

    # Now, remove sources nodes in sequence to get the ranking.
    ranking = []
    scores = []
    alt_idx_remaining = []
    for i in range(m):
      alt_idx_remaining.append(i)
    while len(ranking) < m:
      has_source = False
      for j in range(m):
        if self._is_source(graph, j):
          ranking.append(alternatives[j])
          scores.append(self._get_score(graph, margin_matrix, j))
          self._remove_node(graph, j)
          alt_idx_remaining.remove(j)
          has_source = True
          break
      if not has_source:
        # At the end, it can happen that there are a number of disconnected
        # nodes (no incoming nor outgoing edges). Take the first one from the
        # graph.
        j = alt_idx_remaining[0]
        ranking.append(alternatives[j])
        scores.append(0)
        self._remove_node(graph, j)
        alt_idx_remaining.remove(j)

    # Finally, return the ranking and scores.
    outcome = RankedPairsRankOutcome(
        rankings=ranking, scores=scores, graph=full_graph
    )
    return outcome



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tabular Q-learning agent tests."""

from absl.testing import absltest
import numpy as np

from open_spiel.python import rl_environment
from open_spiel.python import rl_tools
from open_spiel.python.algorithms import tabular_qlearner
import pyspiel
from open_spiel.python.utils import stats

# Fixed seed to make test non stochastic.
SEED = 10000

# A simple two-action game encoded as an EFG game. Going left gets -1, going
# right gets a +1.
SIMPLE_EFG_DATA = """
  EFG 2 R "Simple single-agent problem" { "Player 1" } ""
  p "ROOT" 1 1 "ROOT" { "L" "R" } 0
    t "L" 1 "Outcome L" { -1.0 }
    t "R" 2 "Outcome R" { 1.0 }
"""


def blackjack_info_state_to_string(state):
  if state.is_terminal():
    return str(state)
  else:
    return (
        "Terminal? False\n"
        f"Dealer visible card: {state.dealers_visible_card()}\n"
        f"Player sum: {state.get_best_player_total(0)}\n"
    )


class QlearnerTest(absltest.TestCase):

  def test_simple_game(self):
    game = pyspiel.load_efg_game(SIMPLE_EFG_DATA)
    env = rl_environment.Environment(game=game)

    agent = tabular_qlearner.QLearner(0, game.num_distinct_actions())
    total_reward = 0

    for _ in range(100):
      total_eval_reward = 0
      for _ in range(1000):
        time_step = env.reset()
        while not time_step.last():
          agent_output = agent.step(time_step)
          time_step = env.step([agent_output.action])
          total_reward += time_step.rewards[0]
        agent.step(time_step)
      self.assertGreaterEqual(total_reward, 75)
      for _ in range(1000):
        time_step = env.reset()
        while not time_step.last():
          agent_output = agent.step(time_step, is_evaluation=True)
          time_step = env.step([agent_output.action])
          total_eval_reward += time_step.rewards[0]
      self.assertGreaterEqual(total_eval_reward, 250)

  @absltest.skip("Example when using different info state string.")
  def test_blackjack(self):
    game = pyspiel.load_game("blackjack")
    env = rl_environment.Environment(game=game, include_full_state=True)

    agent = tabular_qlearner.QLearner(
        0,
        game.num_distinct_actions(),
        info_state_to_string_override=blackjack_info_state_to_string,
        epsilon_schedule=rl_tools.ConstantSchedule(0.1),
    )

    train_return_swa = stats.SlidingWindowAccumulator(100000)
    eval_return_swa = stats.SlidingWindowAccumulator(100000)

    for i in range(1000):
      for _ in range(1000):
        episode_return = 0
        time_step = env.reset()
        while not time_step.last():
          agent_output = agent.step(time_step)
          time_step = env.step([agent_output.action])
          episode_return += time_step.rewards[0]
        agent.step(time_step)
        train_return_swa.add(episode_return)
      for _ in range(1000):
        episode_return = 0
        time_step = env.reset()
        while not time_step.last():
          agent_output = agent.step(time_step, is_evaluation=True)
          time_step = env.step([agent_output.action])
          episode_return += time_step.rewards[0]
        eval_return_swa.add(episode_return)
      print(
          f"epoch {i}, avg train return: {train_return_swa.mean()}, "
          + f"avg eval reward: {eval_return_swa.mean()}"
      )


if __name__ == "__main__":
  np.random.seed(SEED)
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest

import jax

from open_spiel.python import rl_agent_policy
from open_spiel.python import rl_environment
from open_spiel.python.jax import boltzmann_dqn
import pyspiel

jax.config.update("jax_threefry_partitionable", False)

# A simple two-action game encoded as an EFG game. Going left gets -1, going
# right gets a +1.
SIMPLE_EFG_DATA = """
  EFG 2 R "Simple single-agent problem" { "Player 1" } ""
  p "ROOT" 1 1 "ROOT" { "L" "R" } 0
    t "L" 1 "Outcome L" { -1.0 }
    t "R" 2 "Outcome R" { 1.0 }
"""


class DQNTest(absltest.TestCase):

  def test_train(self):
    game = pyspiel.load_efg_game(SIMPLE_EFG_DATA)
    env = rl_environment.Environment(game=game)
    agent = boltzmann_dqn.BoltzmannDQN(
        0,
        state_representation_size=game.information_state_tensor_shape()[0],
        num_actions=game.num_distinct_actions(),
        hidden_layers_sizes=[16],
        replay_buffer_capacity=100,
        batch_size=5,
        epsilon_start=0.02,
        epsilon_end=0.01,
        eta=5.0)
    total_reward = 0

    # Training. This will use the epsilon-greedy actions.
    for _ in range(100):
      time_step = env.reset()
      while not time_step.last():
        agent_output = agent.step(time_step)
        time_step = env.step([agent_output.action])
        total_reward += time_step.rewards[0]
      agent.step(time_step)
    self.assertGreaterEqual(total_reward, -100)

    # Update the previous Q-network.
    agent.update_prev_q_network()

    # This will use the soft-max actions.
    policy = rl_agent_policy.RLAgentPolicy(game, agent, 0, False)
    probs = policy.action_probabilities(game.new_initial_state())
    self.assertAlmostEqual(probs[0], 0.54, places=2)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
import jax
import numpy as np

from open_spiel.python import rl_agent
from open_spiel.python.examples.rrps_poprl import impala
from open_spiel.python.examples.rrps_poprl import rl_environment
import pyspiel


# A simple two-action game encoded as an EFG game. Going left gets -1, going
# right gets a +1.
SIMPLE_EFG_DATA = """
  EFG 2 R "Simple single-agent problem" { "Player 1" } ""
  p "ROOT" 1 1 "ROOT" { "L" "R" } 0
    t "L" 1 "Outcome L" { -1.0 }
    t "R" 2 "Outcome R" { 1.0 }
"""


class FixedSequenceAgent(rl_agent.AbstractAgent):
  """An example agent class."""

  def __init__(
      self, player_id, num_actions, sequence, name="fixed_sequence_agent"
  ):
    assert num_actions > 0
    self._player_id = player_id
    self._num_actions = num_actions
    self._sequence = sequence
    self._seq_idx = 0

  def step(self, time_step, is_evaluation=False):
    # If it is the end of the episode, don't select an action.
    if time_step.last():
      return

    probs = np.zeros(self._num_actions)
    action = self._sequence[self._seq_idx]
    self._seq_idx += 1
    if self._seq_idx >= len(self._sequence):
      self._seq_idx = 0
    probs[action] = 1.0

    return rl_agent.StepOutput(action=action, probs=probs)


class IMPALATest(absltest.TestCase):

  def test_simple_game(self):
    game = pyspiel.load_efg_game(SIMPLE_EFG_DATA)
    env = rl_environment.Environment(game=game)
    max_abs_reward = max(
        abs(env.game.min_utility()), abs(env.game.max_utility())
    )
    agent = impala.IMPALA(
        0,
        state_representation_size=game.information_state_tensor_shape()[0],
        num_actions=game.num_distinct_actions(),
        num_players=game.num_players(),
        unroll_len=20,
        net_factory=impala.BasicRNN,
        rng_key=jax.random.PRNGKey(42),
        max_abs_reward=max_abs_reward,
        learning_rate=5e-3,
        hidden_layers_sizes=[16],
        batch_size=5,
    )

    total_reward = 0
    for _ in range(1000):
      time_step = env.reset()
      while not time_step.last():
        agent_output = agent.step(time_step)
        time_step = env.step([agent_output.action])
        total_reward += time_step.rewards[0]
      agent.step(time_step)
    print(total_reward)
    self.assertGreaterEqual(total_reward, 500)

  @absltest.skip("Takes too long to run, but does approach 1.")
  def test_catch(self):
    env = rl_environment.Environment("catch")
    state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]
    num_players = env.num_players
    max_abs_reward = max(
        abs(env.game.min_utility()), abs(env.game.max_utility())
    )
    agent = impala.IMPALA(
        0,
        state_representation_size=state_size,
        num_actions=num_actions,
        num_players=num_players,
        unroll_len=20,
        net_factory=impala.BasicRNN,
        rng_key=jax.random.PRNGKey(42),
        max_abs_reward=max_abs_reward,
        learning_rate=5e-3,
        hidden_layers_sizes=[16],
        batch_size=2,
    )

    window_sum = 0
    window_width = 50
    window_tick = 0
    for ep in range(10000):
      episode_reward = 0
      time_step = env.reset()
      while not time_step.last():
        agent_output = agent.step(time_step)
        time_step = env.step([agent_output.action])
        episode_reward += time_step.rewards[0]
      # print(f"Total reward: {total_reward}")
      # avg_rew = total_reward / (ep + 1)
      agent.step(time_step)
      window_sum += episode_reward
      window_tick += 1
      if window_tick >= window_width:
        avg_window_reward = window_sum / window_width
        window_tick = 0
        window_sum = 0
        print(f"Ep {ep}, avg window rew: {avg_window_reward}")

  @absltest.skip("Takes too long to run, but does approach 1000.")
  def test_run_rps(self):
    env = rl_environment.Environment(
        f"repeated_game(stage_game=matrix_rps(),num_repetitions={pyspiel.ROSHAMBO_NUM_THROWS},recall=1)"
    )
    state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]
    num_players = env.num_players
    max_abs_reward = max(
        abs(env.game.min_utility()), abs(env.game.max_utility())
    )
    agents = [
        impala.IMPALA(  # pylint: disable=g-complex-comprehension
            player_id,
            state_representation_size=state_size,
            num_actions=num_actions,
            num_players=num_players,
            unroll_len=20,
            net_factory=impala.BasicRNN,
            rng_key=jax.random.PRNGKey(seed),
            max_abs_reward=max_abs_reward,
            entropy=0.001,
            learning_rate=0.001,
            hidden_layers_sizes=[64, 32],
            prediction_weight=0,
            discount_factor=0.9,
            batch_size=16,
        )
        for (player_id, seed) in [(0, 238576517), (1, 738328671)]
    ]
    agents[0] = FixedSequenceAgent(0, num_actions, [0, 1, 2, 2, 1, 0])
    for ep in range(1000):
      time_step = env.reset()
      total_rewards = np.zeros(2, dtype=np.float32)
      while not time_step.last():
        agent_outputs = [agents[0].step(time_step), agents[1].step(time_step)]
        actions = [agent_outputs[0].action, agent_outputs[1].action]
        time_step = env.step(actions)
        total_rewards += np.array(time_step.rewards)
      for agent in agents:
        agent.step(time_step)
      print(f"{ep} {total_rewards}")


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Computes a greedy policy from a value."""
import numpy as np

from open_spiel.python import policy as policy_std
from open_spiel.python.mfg import value


class GreedyPolicy(policy_std.Policy):
  """Computes the greedy policy of a value."""

  def __init__(self, game, player_ids, state_action_value: value.ValueFunction):
    """Initializes the greedy policy.

    Args:
      game: The game to analyze.
      player_ids: list of player ids for which this policy applies; each should
        be in the range 0..game.num_players()-1.
      state_action_value: A state-action value function.
    """
    super(GreedyPolicy, self).__init__(game, player_ids)
    self._state_action_value = state_action_value

  def action_probabilities(self, state, player_id=None):
    q = [
        self._state_action_value(state, action)
        for action in state.legal_actions()
    ]
    amax_q = [0.0 for _ in state.legal_actions()]
    amax_q[np.argmax(q)] = 1.0
    return dict(zip(state.legal_actions(), amax_q))

  def action(self, state, player_id=None):
    q = [
        self._state_action_value(state, action)
        for action in state.legal_actions()
    ]
    return state.legal_actions()[np.argmax(q)]



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Outputs value of best response policy against set of distributions."""
import collections
from typing import List
from open_spiel.python.mfg import distribution as distribution_std
from open_spiel.python.mfg import value
import pyspiel


class JointBestResponse(value.ValueFunction):
  """Computes a best response value."""

  def __init__(
      self,
      game,
      distributions: List[distribution_std.Distribution],
      weights,
      root_state=None,
  ):
    """Initializes the joint best response computation.

    The joint best response is computed under the following premisse : the
    player does not know which distribution it is playing against. It only knows
    their probabilities, and thus tries to find a best response against their
    mixture.

    This is accomplished by recursively computing the action that maximizes the
    marginalized value of each node over each distribution.

    Warning : This version only works on games whose observation space &
    dynamics do NOT depend on state distribution.

    Args:
      game: The game to analyze.
      distributions: A list of `distribution_std.Distribution`.
      weights: A list of floats the same length as `distributions`. Represents
        the mixture weight of each member of `distributions`.
      root_state: The state of the game at which to start. If `None`, the game
        root state is used.
    """
    super().__init__(game)
    if root_state is None:
      self._root_states = game.new_initial_states()
    else:
      self._root_states = [root_state]
    self._distributions = distributions
    self._weights = weights
    # Maps states (in string format) to the value of the optimal policy given
    # 'self._distribution'.
    self._state_value = collections.defaultdict(float)
    self.evaluate()

  def get_state_rewards(self, mu_states):
    return sum([
        weight * mu_state.rewards()[mu_state.mean_field_population()]
        for weight, mu_state in zip(self._weights, mu_states)
    ])

  def get_new_mu_states(self, mu_states):
    new_mu_states = []
    for mu_ind, mu_state in enumerate(mu_states):
      dist = [
          self._distributions[mu_ind].value_str(str_state, 0.0)
          for str_state in mu_state.distribution_support()
      ]
      new_mu_state = mu_state.clone()
      new_mu_state.update_distribution(dist)
      new_mu_states.append(new_mu_state)
    return new_mu_states

  def eval_state(self, mu_states):
    """Evaluate the value of a state.

    Args:
      mu_states: A list of game states, one for each `distributions` member.

    Returns:
      The optimal value of the state.

    Recursively computes the value of the optimal policy given the fixed state
    distributions. `self._state_value` is used as a cache for pre-computed
    values.
    """
    state = mu_states[0]
    state_str = state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID)
    if state_str in self._state_value:
      return self._state_value[state_str]
    if state.is_terminal():
      self._state_value[state_str] = self.get_state_rewards(mu_states)
      return self._state_value[state_str]
    if state.current_player() == pyspiel.PlayerId.CHANCE:
      self._state_value[state_str] = 0.0
      for action, prob in state.chance_outcomes():
        new_mu_states = [mu_state.child(action) for mu_state in mu_states]
        self._state_value[state_str] += prob * self.eval_state(new_mu_states)
      return self._state_value[state_str]
    if state.current_player() == pyspiel.PlayerId.MEAN_FIELD:
      new_mu_states = self.get_new_mu_states(mu_states)
      self._state_value[state_str] = self.get_state_rewards(
          mu_states
      ) + self.eval_state(new_mu_states)
      return self._state_value[state_str]
    else:
      assert int(state.current_player()) >= 0, "The player id should be >= 0"
      max_q = max(
          self.eval_state([mu_state.child(action) for mu_state in mu_states])
          for action in state.legal_actions()
      )
      self._state_value[state_str] = self.get_state_rewards(mu_states) + max_q
      return self._state_value[state_str]

  def evaluate(self):
    """Evaluate the best response value on all states."""
    for state in self._root_states:
      self.eval_state([state.clone() for _ in self._distributions])

  def value(self, state, action=None):
    if action is None:
      return self._state_value[state.observation_string(
          pyspiel.PlayerId.DEFAULT_PLAYER_ID)]
    new_state = state.child(action)
    return state.rewards()[state.mean_field_population()] + self._state_value[
        new_state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID)]



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Runs Deep Average-network Fictitious Play with DQN agents."""

import os
from typing import Sequence

from absl import flags
import jax

from open_spiel.python import policy as policy_std
from open_spiel.python import rl_environment
from open_spiel.python.jax import dqn
from open_spiel.python.mfg import utils
from open_spiel.python.mfg.algorithms import average_network_fictitious_play
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.games import factory
from open_spiel.python.utils import app
from open_spiel.python.utils import metrics
from open_spiel.python.utils import training

_GAME_NAME = flags.DEFINE_string('game_name', 'mfg_crowd_modelling_2d',
                                 'Name of the game.')
_ENV_SETTING = flags.DEFINE_string(
    'env_setting', 'mfg_crowd_modelling_2d',
    'Name of the game settings. If None, the game name will be used.')
_LOGDIR = flags.DEFINE_string(
    'logdir', None,
    'Logging dir to use for TF summary files. If None, the metrics will only '
    'be logged to stderr.')
_LOG_DISTRIBUTION = flags.DEFINE_bool('log_distribution', False,
                                      'Enables logging of the distribution.')
_NUM_ITERATIONS = flags.DEFINE_integer('num_iterations', 100,
                                       'Number of iterations.')
_EVAL_EVERY = flags.DEFINE_integer(
    'eval_every', 200, 'Episode frequency at which the agents are evaluated.')

# Flags for best response RL (DQN) agent.
# Training options.
_BATCH_SIZE = flags.DEFINE_integer(
    'batch_size', 128, 'Number of transitions to sample at each learning step.')
_LEARN_EVERY = flags.DEFINE_integer(
    'learn_every', 40, 'Number of steps between learning updates.')
_NUM_DQN_EPISODES_PER_ITERATION = flags.DEFINE_integer(
    'num_dqn_episodes_per_iteration', 3000,
    'Number of DQN training episodes for each iteration.')
_EPSILON_DECAY_DURATION = flags.DEFINE_integer(
    'epsilon_decay_duration', int(20e6),
    'Number of game steps over which epsilon is decayed.')
_EPSILON_START = flags.DEFINE_float('epsilon_start', 0.1,
                                    'Starting exploration parameter.')
_EPSILON_END = flags.DEFINE_float('epsilon_end', 0.1,
                                  'Final exploration parameter.')
_DISCOUNT_FACTOR = flags.DEFINE_float('discount_factor', 1.0,
                                      'Discount factor for future rewards.')
_SEED = flags.DEFINE_integer('seed', 42, 'Training seed.')
# Network options.
_HIDDEN_LAYERS_SIZES = flags.DEFINE_list(
    'hidden_layers_sizes', ['128', '128'],
    'Number of hidden units in the Q-net.')
_UPDATE_TARGET_NETWORK_EVERY = flags.DEFINE_integer(
    'update_target_network_every', 200,
    'Number of steps between DQN target network updates.')
# Replay buffer options.
_REPLAY_BUFFER_CAPACITY = flags.DEFINE_integer('replay_buffer_capacity', 5000,
                                               'Size of the replay buffer.')
_MIN_BUFFER_SIZE_TO_LEARN = flags.DEFINE_integer(
    'min_buffer_size_to_learn', 200,
    'Number of samples in buffer before learning begins.')
# Loss and optimizer options.
_OPTIMIZER = flags.DEFINE_enum('optimizer', 'adam', ['sgd', 'adam'],
                               'Optimizer.')
_LEARNING_RATE = flags.DEFINE_float('learning_rate', 0.001,
                                    'Learning rate for inner rl agent.')
_LOSS = flags.DEFINE_enum('loss', 'mse', ['mse', 'huber'], 'Loss function.')
_HUBER_LOSS_PARAMETER = flags.DEFINE_float('huber_loss_parameter', 1.0,
                                           'Parameter for Huber loss.')
_GRADIENT_CLIPPING = flags.DEFINE_float('gradient_clipping', 40,
                                        'Value to clip the gradient to.')

# Flags for average policy RL agent.
# Training options.
_AVG_POL_BATCH_SIZE = flags.DEFINE_integer(
    'avg_pol_batch_size', 128,
    'Number of transitions to sample at each learning step.')
_AVG_POL_NUM_TRAINING_STEPS_PER_ITERATION = flags.DEFINE_integer(
    'avg_pol_num_training_steps_per_iteration', 2000,
    'Number of steps for average policy at each FP iteration.')
_AVG_POL_NUM_EPISODES_PER_ITERATION = flags.DEFINE_integer(
    'avg_pol_num_episodes_per_iteration', 100,
    'Number of samples to store at each FP iteration.')
# Network options.
_AVG_POL_HIDDEN_LAYERS_SIZES = flags.DEFINE_list(
    'avg_pol_hidden_layers_sizes', ['128', '128'],
    'Number of hidden units in the avg-net and Q-net.')
# Reservoir buffer options.
_AVG_POL_RESERVOIR_BUFFER_CAPACITY = flags.DEFINE_integer(
    'avg_pol_reservoir_buffer_capacity', 100000000,
    'Size of the reservoir buffer.')
_AVG_POL_MIN_BUFFER_SIZE_TO_LEARN = flags.DEFINE_integer(
    'avg_pol_min_buffer_size_to_learn', 100,
    'Number of samples in buffer before learning begins.')
# Loss and optimizer options.
_AVG_POL_OPTIMIZER = flags.DEFINE_enum('avg_pol_optimizer', 'sgd',
                                       ['sgd', 'adam'], 'Optimizer.')
_AVG_POL_LEARNING_RATE = flags.DEFINE_float(
    'avg_pol_learning_rate', 0.01, 'Learning rate for inner rl agent.')
_AVG_GRADIENT_CLIPPING = flags.DEFINE_float('avg_gradient_clipping', 100,
                                            'Value to clip the gradient to.')
_AVG_POL_TAU = flags.DEFINE_float('avg_pol_tau', 10.0,
                                  'Temperature for softmax in policy.')


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  game = factory.create_game_with_setting(_GAME_NAME.value, _ENV_SETTING.value)
  num_players = game.num_players()

  # Create the environments with uniform initial policy.
  uniform_policy = policy_std.UniformRandomPolicy(game)
  uniform_dist = distribution.DistributionPolicy(game, uniform_policy)

  envs = [
      rl_environment.Environment(
          game, mfg_distribution=uniform_dist, mfg_population=p)
      for p in range(num_players)
  ]

  env = envs[0]
  info_state_size = env.observation_spec()['info_state'][0]
  num_actions = env.action_spec()['num_actions']

  # Best response policy agents.
  kwargs_dqn = {
      'batch_size': _BATCH_SIZE.value,
      'discount_factor': _DISCOUNT_FACTOR.value,
      'epsilon_decay_duration': _EPSILON_DECAY_DURATION.value,
      'epsilon_end': _EPSILON_END.value,
      'epsilon_start': _EPSILON_START.value,
      'gradient_clipping': _GRADIENT_CLIPPING.value,
      'hidden_layers_sizes': [int(l) for l in _HIDDEN_LAYERS_SIZES.value],
      'huber_loss_parameter': _HUBER_LOSS_PARAMETER.value,
      'learn_every': _LEARN_EVERY.value,
      'learning_rate': _LEARNING_RATE.value,
      'loss_str': _LOSS.value,
      'min_buffer_size_to_learn': _MIN_BUFFER_SIZE_TO_LEARN.value,
      'optimizer_str': _OPTIMIZER.value,
      'replay_buffer_capacity': _REPLAY_BUFFER_CAPACITY.value,
      'seed': _SEED.value,
      'update_target_network_every': _UPDATE_TARGET_NETWORK_EVERY.value,
  }
  br_rl_agents = [
      dqn.DQN(p, info_state_size, num_actions, **kwargs_dqn)
      for p in range(num_players)
  ]

  num_training_steps_per_iteration = (
      _AVG_POL_NUM_TRAINING_STEPS_PER_ITERATION.value)

  # Metrics writer will also log the metrics to stderr.
  just_logging = _LOGDIR.value is None or jax.host_id() > 0
  writer = metrics.create_default_writer(
      _LOGDIR.value, just_logging=just_logging)

  def logging_fn(it, step, vals):
    writer.write_scalars(it * num_training_steps_per_iteration + step, vals)

  # Average policy agents.
  kwargs_avg = {
      'batch_size': _AVG_POL_BATCH_SIZE.value,
      'hidden_layers_sizes': [
          int(l) for l in _AVG_POL_HIDDEN_LAYERS_SIZES.value
      ],
      'reservoir_buffer_capacity': _AVG_POL_RESERVOIR_BUFFER_CAPACITY.value,
      'learning_rate': _AVG_POL_LEARNING_RATE.value,
      'min_buffer_size_to_learn': _AVG_POL_MIN_BUFFER_SIZE_TO_LEARN.value,
      'optimizer_str': _AVG_POL_OPTIMIZER.value,
      'gradient_clipping': _AVG_GRADIENT_CLIPPING.value,
      'seed': _SEED.value,
      'tau': _AVG_POL_TAU.value
  }
  fp = average_network_fictitious_play.AverageNetworkFictitiousPlay(
      game,
      envs,
      br_rl_agents,
      _AVG_POL_NUM_EPISODES_PER_ITERATION.value,
      num_training_steps_per_iteration,
      eval_every=_EVAL_EVERY.value,
      logging_fn=logging_fn,
      **kwargs_avg)

  def log_metrics(it):
    """Logs the training metrics for each iteration."""
    initial_states = game.new_initial_states()
    distrib = distribution.DistributionPolicy(game, fp.policy)
    pi_value = policy_value.PolicyValue(game, distrib, fp.policy)
    m = {
        f'best_response/{state}': pi_value.eval_state(state)
        for state in initial_states
    }
    m.update({
        f'br_agent{i}/loss': agent.loss for i, agent in enumerate(br_rl_agents)
    })
    nash_conv_fp = nash_conv.NashConv(game, fp.policy)
    m['nash_conv_fp'] = nash_conv_fp.nash_conv()
    logging_fn(it, 0, m)

    # Also save the distribution.
    if _LOG_DISTRIBUTION.value and not just_logging:
      filename = os.path.join(_LOGDIR.value, f'distribution_{it}.pkl')
      utils.save_parametric_distribution(nash_conv_fp.distribution, filename)

  for it in range(_NUM_ITERATIONS.value):
    # Train the RL agent to learn a best response.
    training.run_episodes(
        envs,
        br_rl_agents,
        num_episodes=_NUM_DQN_EPISODES_PER_ITERATION.value,
        is_evaluation=False)

    # Run an iteration of average-network fictitious play and log the metrics.
    fp.iteration()
    log_metrics(it + 1)

  # Make sure all values were written.
  writer.flush()


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.plurality."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.voting import base
from open_spiel.python.voting import plurality

_SIMPLE_VOTE = [["a", "b", "c"], ["a", "c", "b"], ["b", "a", "c"]]
_SIMPLE_WINNER = (_SIMPLE_VOTE, "a")
_WEIGHTED_WINNER = (_SIMPLE_VOTE, [1, 2, 3], [3, 3, 0], ["a", "b"])


class PluralityVotingTest(parameterized.TestCase):
  def setUp(self):
    super().setUp()
    self.method = plurality.PluralityVoting()

  @parameterized.parameters(_SIMPLE_WINNER)
  def test_plurality_with_votes_in_profile_constructor(self, votes, winner):
    profile = base.PreferenceProfile(votes=votes)
    outcome = self.method.run_election(profile)
    self.assertEqual(outcome.ranking[0], winner)

  @parameterized.parameters(_SIMPLE_WINNER)
  def test_plurality_with_alternatives_specified(self, votes, winner):
    profile = base.PreferenceProfile(alternatives=["c", "b", "a"])
    for vote in votes:
      profile.add_vote(vote)
    outcome = self.method.run_election(profile)
    self.assertEqual(outcome.ranking[0], winner)

  @parameterized.parameters(_SIMPLE_WINNER)
  def test_plurality_with_no_default_votes(self, votes, winner):
    profile = base.PreferenceProfile()
    for vote in votes:
      profile.add_vote(vote)
    outcome = self.method.run_election(profile)
    self.assertEqual(outcome.ranking[0], winner)

  @parameterized.parameters(_WEIGHTED_WINNER)
  def test_plurality_with_weighted_votes(self, votes, weights,
                                         correct_scores, winner):
    profile = base.PreferenceProfile()
    for i, vote in enumerate(votes):
      profile.add_vote(vote, weight=weights[i])
    outcome = self.method.run_election(profile)

    with self.subTest("Weighted score correctly calculated."):
      self.assertListEqual(correct_scores, outcome.scores)
    with self.subTest("Winners take the top spots in the ranking."):
      self.assertCountEqual(outcome.ranking[:len(winner)], winner)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python base module for the implementations of Monte Carlo Counterfactual Regret Minimization."""

import numpy as np
from open_spiel.python import policy

REGRET_INDEX = 0
AVG_POLICY_INDEX = 1


class AveragePolicy(policy.Policy):
  """A policy object representing the average policy for MCCFR algorithms."""

  def __init__(self, game, player_ids, infostates):
    # Do not create a copy of the dictionary
    # but work on the same object
    super().__init__(game, player_ids)
    self._infostates = infostates

  def action_probabilities(self, state, player_id=None):
    """Returns the MCCFR average policy for a player in a state.

    If the policy is not defined for the provided state, a uniform
    random policy is returned.

    Args:
      state: A `pyspiel.State` object.
      player_id: Optional, the player id for which we want an action. Optional
        unless this is a simultaneous state at which multiple players can act.

    Returns:
      A `dict` of `{action: probability}` for the specified player in the
      supplied state. If the policy is defined for the state, this
      will contain the average MCCFR strategy defined for that state.
      Otherwise, it will contain all legal actions, each with the same
      probability, equal to 1 / num_legal_actions.
    """
    if player_id is None:
      player_id = state.current_player()
    legal_actions = state.legal_actions()
    info_state_key = state.information_state_string(player_id)
    retrieved_infostate = self._infostates.get(info_state_key, None)
    if retrieved_infostate is None:
      return {a: 1 / len(legal_actions) for a in legal_actions}
    avstrat = (
        retrieved_infostate[AVG_POLICY_INDEX] /
        retrieved_infostate[AVG_POLICY_INDEX].sum())
    return {legal_actions[i]: avstrat[i] for i in range(len(legal_actions))}


class MCCFRSolverBase(object):
  """A base class for both outcome MCCFR and external MCCFR."""

  def __init__(self, game):
    self._game = game
    self._infostates = {}  # infostate keys -> [regrets, avg strat]
    self._num_players = game.num_players()

  def _lookup_infostate_info(self, info_state_key, num_legal_actions):
    """Looks up an information set table for the given key.

    Args:
      info_state_key: information state key (string identifier).
      num_legal_actions: number of legal actions at this information state.

    Returns:
      A list of:
        - the average regrets as a numpy array of shape [num_legal_actions]
        - the average strategy as a numpy array of shape
        [num_legal_actions].
          The average is weighted using `my_reach`
    """
    retrieved_infostate = self._infostates.get(info_state_key, None)
    if retrieved_infostate is not None:
      return retrieved_infostate

    # Start with a small amount of regret and total accumulation, to give a
    # uniform policy: this will get erased fast.
    self._infostates[info_state_key] = [
        np.ones(num_legal_actions, dtype=np.float64) / 1e6,
        np.ones(num_legal_actions, dtype=np.float64) / 1e6,
    ]
    return self._infostates[info_state_key]

  def _add_regret(self, info_state_key, action_idx, amount):
    self._infostates[info_state_key][REGRET_INDEX][action_idx] += amount

  def _add_avstrat(self, info_state_key, action_idx, amount):
    self._infostates[info_state_key][AVG_POLICY_INDEX][action_idx] += amount

  def average_policy(self):
    """Computes the average policy, containing the policy for all players.

    Returns:
      An average policy instance that should only be used during
      the lifetime of solver object.
    """
    return AveragePolicy(self._game, list(range(self._num_players)),
                         self._infostates)

  def _regret_matching(self, regrets, num_legal_actions):
    """Applies regret matching to get a policy.

    Args:
      regrets: numpy array of regrets for each action.
      num_legal_actions: number of legal actions at this state.

    Returns:
      numpy array of the policy indexed by the index of legal action in the
      list.
    """
    positive_regrets = np.maximum(regrets,
                                  np.zeros(num_legal_actions, dtype=np.float64))
    sum_pos_regret = positive_regrets.sum()
    if sum_pos_regret <= 0:
      return np.ones(num_legal_actions, dtype=np.float64) / num_legal_actions
    else:
      return positive_regrets / sum_pos_regret



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Examples of schedule negotations -- useful for generating more examples.
"""

from open_spiel.python.games.chat_games.envs.observations import summary
from open_spiel.python.games.chat_games.envs.termination import utils as term_utils
from open_spiel.python.games.chat_games.envs.utils import text

# Scenario A
OOO_LIST_A = ['monday: false',
              'tuesday: true',
              'wednesday: true',
              'thursday: false',
              'friday: false',
              'saturday: true',
              'sunday: false']
OOO_A = '\n'.join(text.wrap(OOO_LIST_A))

DAY_PREFS_LIST_A = ['monday: 10',
                    'tuesday: 5',
                    'wednesday: 15',
                    'thursday: 3',
                    'friday: 2',
                    'saturday: 1',
                    'sunday: 1'
                    ]
DAY_PREFS_A = '\n'.join(text.wrap(DAY_PREFS_LIST_A))

SCENARIO_A_LIST = ['Hi {receiver},',
                   'I would like to propose meeting on thursday.',
                   'Would you like to meet with me then?',
                   'Best,', '{sender}']
SCENARIO_A = '\n\n'.join(text.wrap(SCENARIO_A_LIST))

# Scenario B
OOO_LIST_B = ['monday: true',
              'tuesday: false',
              'wednesday: true',
              'thursday: false',
              'friday: false',
              'saturday: true',
              'sunday: false']
OOO_B = '\n'.join(text.wrap(OOO_LIST_B))

DAY_PREFS_LIST_B = ['monday: 5',
                    'tuesday: 5',
                    'wednesday: 5',
                    'thursday: 1',
                    'friday: 1',
                    'saturday: 1',
                    'sunday: 1'
                    ]
DAY_PREFS_B = '\n'.join(text.wrap(DAY_PREFS_LIST_B))

SCENARIO_B_LIST = ['Hi {receiver},',
                   'I strongly urge you to meet me on friday when I am in ' +
                   'the office.',
                   'what do you say?',
                   'Best,', '{sender}']
SCENARIO_B = '\n\n'.join(text.wrap(SCENARIO_B_LIST))

query = ('Read the following summary of a dialgoue between two parties ' +
         'attempting to reach an agreement. Have the players reached an ' +
         'agreement? If a meeting time has been accepted or the players ' +
         'cannot come to an agreement, respond Yes. Otherwise, if the ' +
         'players are still discussing terms, respond No.' +
         'Here is the dialogue:\n\n{msg}\n\n' + '&' *50 +
         '\n\nHave all parties agreed on a meeting time?'
         '\nResponse: ')

LLM_TERMINATION_PROMPT = term_utils.Termination(query,
                                                summary.PREFIX,
                                                summary.POSTFIX)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""WoLF policy-hill climbing agent.

Based on: https://www.sciencedirect.com/science/article/pii/S0004370202001212
"""

import collections
import numpy as np

from open_spiel.python import rl_agent
from open_spiel.python import rl_tools
from open_spiel.python.algorithms.projected_replicator_dynamics import _simplex_projection


def valuedict():
  return collections.defaultdict(float)


class WoLFSchedule(rl_tools.ValueSchedule):
  """Schedule rules described in the WoLF paper.

  at step t the step size is (t0 / (t + t1))
  """

  def __init__(self, t0, t1):
    super(WoLFSchedule, self).__init__()
    self._t0 = t0
    self._t1 = t1
    self._step_taken = 0

  def step(self):
    value = (self._t0 / (self._step_taken + self._t1))
    self._step_taken += 1
    return value

  @property
  def value(self):
    return self._t0 / (self._step_taken + self._t1)


class WoLFPHC(rl_agent.AbstractAgent):
  """WoLF policy-hill climbing agent agent.


    Based on win or learn fast principle.
    Based on:
    https://www.sciencedirect.com/science/article/pii/S0004370202001212
  """

  def __init__(self,
               player_id,
               num_actions,
               step_size=WoLFSchedule(10000, 1000000),
               epsilon_schedule=rl_tools.ConstantSchedule(0.2),
               delta_w=WoLFSchedule(1, 20000),
               delta_l=WoLFSchedule(2, 20000),
               discount_factor=1.0):
    """Initialize the WoLF-PHC agent."""
    self._player_id = player_id
    self._num_actions = num_actions
    self._step_size = step_size
    self._epsilon_schedule = epsilon_schedule
    self._epsilon = epsilon_schedule.value
    self._discount_factor = discount_factor
    self._delta_w = delta_w
    self._delta_l = delta_l
    self._cur_policy = collections.defaultdict(valuedict)
    self._avg_policy = collections.defaultdict(valuedict)
    self._q_values = collections.defaultdict(valuedict)
    self._state_counters = valuedict()
    self._prev_info_state = None
    self._last_loss_value = None
    self._cur_delta_value = self._delta_l.value

  def _hill_climbing(self, info_state, legal_actions):
    """Does the hill-climbing update.

    Args:
      info_state: hashable representation of the information state.
      legal_actions: list of actions at `info_state`.
    """

    greedy_q = max(
        [self._q_values[info_state][action] for action in legal_actions])
    greedy_actions = [
        action for action in legal_actions
        if self._q_values[info_state][action] == greedy_q
    ]
    if len(greedy_actions) == len(legal_actions):
      return

    deltas = {  # pylint: disable=g-complex-comprehension
        action:
        min(self._cur_policy[info_state][action],
            self._cur_delta_value / (len(legal_actions) - len(greedy_actions)))
        for action in legal_actions
    }

    delta_greedy = sum([
        deltas[action]
        for action in legal_actions
        if action not in greedy_actions
    ]) / len(greedy_actions)

    deltas = {
        action:
        -deltas[action] if action not in greedy_actions else delta_greedy
        for action in legal_actions
    }
    new_policy = np.array([
        self._cur_policy[info_state][action] + deltas[action]
        for action in legal_actions
    ])
    new_policy = _simplex_projection(new_policy)
    for i in range(len(legal_actions)):
      self._cur_policy[info_state][legal_actions[i]] = new_policy[i]

  def _get_action_probs(self, info_state, legal_actions, epsilon):
    """Returns a selected action and the probabilities of legal actions.

    To be overwritten by subclasses that implement other action selection
    methods.
    Args:
      info_state: hashable representation of the information state.
      legal_actions: list of actions at `info_state`.
      epsilon: float: current value of the epsilon schedule or 0 in case
        evaluation. QLearner uses it as the exploration parameter in
        epsilon-greedy, but subclasses are free to interpret in different ways
        (e.g. as temperature in softmax).
    """
    if info_state not in self._cur_policy:
      for action in legal_actions:
        self._cur_policy[info_state][action] = 1. / len(legal_actions)
        self._avg_policy[info_state][action] = 1. / len(legal_actions)

    probs = np.zeros(self._num_actions)
    for action in legal_actions:
      probs[action] = ((1-epsilon) * self._cur_policy[info_state][action] +
                       epsilon * 1.0 / len(legal_actions))
    action = np.random.choice(range(self._num_actions), p=probs)
    return action, probs

  def step(self, time_step, is_evaluation=False):
    """Returns the action to be taken and updates the Q-values if needed.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """

    info_state = str(time_step.observations["info_state"][self._player_id])
    legal_actions = time_step.observations["legal_actions"][self._player_id]

    # Prevent undefined errors if this agent never plays until terminal step
    action, probs = None, None

    # Act step: don't act at terminal states.
    if not time_step.last():
      epsilon = 0.0 if is_evaluation else self._epsilon
      action, probs = self._get_action_probs(info_state, legal_actions, epsilon)

    # Learn step: don't learn during evaluation or at first agent steps.
    if self._prev_info_state and not is_evaluation:
      target = time_step.rewards[self._player_id]
      if not time_step.last():  # Q values are zero for terminal.
        target += self._discount_factor * max(
            [self._q_values[info_state][a] for a in legal_actions])

      prev_q_value = self._q_values[self._prev_info_state][self._prev_action]
      self._last_loss_value = target - prev_q_value
      self._q_values[self._prev_info_state][self._prev_action] += (
          self._step_size.value * self._last_loss_value)

      self._state_counters[info_state] += 1
      for action_ in legal_actions:
        self._avg_policy[info_state][action_] = (
            self._avg_policy[info_state][action_] +
            1 / self._state_counters[info_state] * (
                self._cur_policy[info_state][action_] -
                self._avg_policy[info_state][action_]))

      assert self._delta_l.value > self._delta_w.value
      cur_policy_value = sum([
          self._cur_policy[info_state][action] *
          self._q_values[info_state][action] for action in legal_actions
      ])
      avg_policy_value = sum([
          self._avg_policy[info_state][action] *
          self._q_values[info_state][action] for action in legal_actions
      ])
      if cur_policy_value > avg_policy_value:
        self._cur_delta_value = self._delta_w.value
      else:
        self._cur_delta_value = self._delta_l.value

      if not time_step.last():
        self._hill_climbing(info_state, legal_actions)

        # Decay epsilon, if necessary.
        self._epsilon = self._epsilon_schedule.step()
        self._delta_l.step()
        self._delta_w.step()
        self._step_size.step()
      else:  # prepare for the next episode.
        self._prev_info_state = None
        return

    # Don't mess up with the state during evaluation.
    if not is_evaluation:
      self._prev_info_state = info_state
      self._prev_action = action
    return rl_agent.StepOutput(action=action, probs=probs)

  @property
  def loss(self):
    return self._last_loss_value



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for regret minimizers."""

import numpy as np

from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import utils
from open_spiel.python.mfg.algorithms.regret import c_ce_optimization


class RegretMinimizer(object):
  """Base class for Regret Minimizers.

  Implements base functions for regret minimizers to implement.

  Attributes:
    _game: Pyspiel game.
    _regret_steps_per_step: Number of regret steps per `step` call (Maximum
      number in case `stop_early` is true)
    _rho_tol: If `_compress_nus` is true, minimum probability threshold (
      Probabilities below `rho_tol` will be filtered out).
    _compress_nus: Whether to compress nus (Remove nus with low selection
      probability) or not.
    _compress_lbd: Penalty term in L1 minimization when compressing nus.
    _stop_early: Whether to stop regret computation when average regret is lower
      than `_stop_regret_threshold` or to keep going until
      `_regret_steps_per_step` steps have been accomplished.
    _stop_regret_threshold: If `stop_early` is true, average regret threshold
      under which the algorithm will stop.
    _policies: List of Policies
    _value_estimator: Value estimation function.
    _value_estimation_n: Number of runs to average _value_estimator's result on.
  """

  def __init__(
      self,
      game,
      policies,
      regret_steps_per_step: int = 1,
      rho_tol: float = 1e-4,
      compress_nus: bool = True,
      compress_every: int = 1,
      compress_lbd: float = 0.0,
      stop_early: bool = True,
      stop_regret_threshold: float = 1e-3,
      value_estimator=utils.sample_value,
      value_estimation_n: int = 1,
      compute_internal_regret: bool = False,
  ):
    self._game = game
    self._regret_steps_per_step = regret_steps_per_step

    self._compress_nus = compress_nus
    self._compress_every = compress_every
    self._compress_lbd = compress_lbd

    self._stop_early = stop_early
    self._stop_regret_threshold = stop_regret_threshold

    self._rho_tol = rho_tol
    self._policies = policies

    self._value_estimator = value_estimator
    self._value_estimation_n = value_estimation_n

    self._compute_internal_regret = compute_internal_regret

    self._nus = []
    self._rewards = []
    self._true_rewards = []
    self._policy_mus = []
    self._nu_weights = []

  def update_policy_mus(self):
    """Update the stored distributions of our policies."""
    self._policy_mus = [
        distribution.DistributionPolicy(self._game, policy)
        for policy in self._policies
    ]

  def get_nu(self):
    """Returns current Population Distribution."""
    raise NotImplementedError

  def step(self, welfare_bonus=0.0):
    raise NotImplementedError

  def step_for(
      self,
      T,  # pylint: disable=invalid-name
      initial_welfare_bonus=None,
      welfare_decay=None,
      use_true_rewards_when_compressing=True,
  ):
    """Call `step` method `T` times maximum, potentially stop early.

    Args:
      T: Maximum number of `step` calls to run.
      initial_welfare_bonus: How much to initially reward high-welfare-inducing
        actions.
      welfare_decay: Welfare decay term.
      use_true_rewards_when_compressing: Compress and compute optimal (C)CE
        according to true rewards (= True) or according to modified rewards (=
        False)
    """
    welfare_bonus = 0.0
    if initial_welfare_bonus is not None:
      assert welfare_decay is not None
      welfare_bonus = initial_welfare_bonus

    weights = None
    for t in range(T):
      if welfare_decay is not None:
        welfare_bonus = max(0.0, welfare_bonus - welfare_decay * t / T)
      self.step(welfare_bonus=welfare_bonus)
      if self._stop_early and (t % self._compress_every == 0):
        try:
          regret, weights = self.get_post_compression_regret_and_weights(
              use_true_rewards_when_compressing=use_true_rewards_when_compressing
          )
          # print("\t\t{}".format(regret))
          assert np.abs(np.sum(weights) - 1.0) < 1e-8, np.sum(weights)
        except:  # pylint: disable=bare-except
          print("Simplex method encountered an error.")
          continue
        if regret < self._stop_regret_threshold:
          break
    if weights is None and self._compress_nus:
      regret, weights = self.get_post_compression_regret_and_weights(
          use_true_rewards_when_compressing=use_true_rewards_when_compressing
      )
    if self._compress_nus:
      self.compress_nus_and_weights(weights)

  def get_post_compression_regret_and_weights(
      self, use_true_rewards_when_compressing=True
  ):
    """Computes optimized (C)CE by varying the temporal weight on each `nu`.

    Args:
      use_true_rewards_when_compressing: compute optimal (C)CE according to true
        rewards (= True) or according to modified rewards (= False)

    Returns:
      Regret for new temporal weights, new temporal weights
    """
    if self._compute_internal_regret:
      nu_weights = c_ce_optimization.compress_internal_weights(
          self.get_nus(),
          self.compute_regrets(
              use_true_rewards=use_true_rewards_when_compressing
          ),
          rewards=self._rewards,
          lbd=self._compress_lbd,
      )
      regret = np.max([
          np.max(np.sum(nu_weights.reshape(-1, 1) * a, axis=0))
          for a in self.compute_regrets(
              use_true_rewards=use_true_rewards_when_compressing
          )
      ])
    else:
      nu_weights = c_ce_optimization.compress_external_weights(
          self.get_nus(),
          self.compute_regrets(
              use_true_rewards=use_true_rewards_when_compressing
          ),
          rewards=self._rewards,
          lbd=self._compress_lbd,
      )
      regret = np.max(
          np.sum(
              nu_weights.reshape(-1, 1)
              * self.compute_regrets(
                  use_true_rewards=use_true_rewards_when_compressing
              ),
              axis=0,
          )
      )
    return regret, nu_weights

  def compress_nus_and_weights(self, nu_weights):
    """Run L1 optimization to only keep important members of `nus`."""
    if self._compress_nus:
      if np.abs(np.sum(nu_weights) - 1.0) > 1e-8:
        # If the optimization was unsuccessful, do *not* compress.
        print(
            "Unsuccessful optimization, weights sum to {}".format(
                np.sum(nu_weights)
            )
        )
        return
      new_nus = [
          nu
          for weight, nu in zip(nu_weights, self._nus)
          if weight > self._rho_tol
      ]
      new_rewards = [
          reward
          for weight, reward in zip(nu_weights, self._rewards)
          if weight > self._rho_tol
      ]
      new_true_rewards = [
          reward
          for weight, reward in zip(nu_weights, self._true_rewards)
          if weight > self._rho_tol
      ]

      new_nu_weights = [
          weight for weight in nu_weights if weight > self._rho_tol
      ]
      new_nu_weights = np.array(new_nu_weights) / np.sum(new_nu_weights)

      self._nus = new_nus
      self._rewards = new_rewards
      self._true_rewards = new_true_rewards
      self._nu_weights = new_nu_weights

  def reset(self, policies):
    """Restart the bandit with new policies."""
    raise NotImplementedError

  def increase_precision_x_fold(self, x):
    self._stop_regret_threshold /= x
    self._rho_tol /= x
    self._regret_steps_per_step *= x

  def compute_p(self):
    """Computes `p` as presented in Blum's External to Internal Regret."""
    assert (
        self._compute_internal_regret
    ), "`p` does not exist when computing external regret."
    w_nus = np.array(self.get_all_w_nus())

    p = np.ones(len(self._policies))
    pprime = np.dot(p, w_nus)
    n_trials = 100000
    i = 0
    while np.sum(np.abs(pprime - p)) > 1e-8 and i < n_trials:
      p = pprime
      pprime = np.dot(p, w_nus)
      i += 1

    if np.sum(np.abs(pprime - p)) > 1e-8 and i >= n_trials:
      raise ValueError(
          "Power method did not converge after {} trials.".format(n_trials)
      )
    self._p = p / np.sum(p)

  def get_all_w_nus(self):
    """returns all nus for all times and all policies."""
    raise NotImplementedError

  def compute_regrets(self, use_true_rewards=False):
    """Computes the algorithm's current external/internal regrets.

    Args:
      use_true_rewards: Whether to use altered game rewards, or true game
        rewards.

    Returns:
      Internal regret of shape [T, P, P] if `self._compute_internal_regret` is
        true, otherwise external regret of shape [T, P], where T is the current
        number of iterations and P the number of policies.
    """
    if use_true_rewards:
      rewards = self._true_rewards
    else:
      rewards = self._rewards

    if self._compute_internal_regret:
      regrets = []
      nus = np.array(self._nus)
      rewards = np.array(rewards)
      for action in range(rewards.shape[1]):
        on_policy_values = (rewards[:, action] * nus[:, action]).reshape(-1, 1)
        action_values = rewards * nus[:, action].reshape(-1, 1)
        regrets.append(action_values - on_policy_values)
    else:
      on_policy_value = np.sum(
          rewards * np.array(self._nus), axis=1, keepdims=True
      )
      policy_value = rewards
      regrets = policy_value - on_policy_value
    return regrets

  def compute_average_regret(self, use_true_rewards=True):
    """Computes the algorithm's average external/internal regrets.

    Args:
      use_true_rewards: Whether to use altered game rewards, or true game
        rewards.

    Returns:
      Internal regret if `self._compute_internal_regret` is true, otherwise
        external regret.
    """

    if use_true_rewards:
      rewards = self._true_rewards
    else:
      rewards = self._rewards

    nu_weights = self.get_normalized_nu_weights()
    if self._compute_internal_regret:
      regrets = 0.0
      nus = np.array(self._nus)
      rewards = np.array(rewards)
      for action in range(rewards.shape[1]):
        on_policy_values = (rewards[:, action] * nus[:, action]).reshape(-1, 1)
        action_values = rewards * nus[:, action].reshape(-1, 1)
        regrets += np.max(
            np.sum(
                nu_weights.reshape(-1, 1) * (action_values - on_policy_values),
                axis=0,
            )
        )
    else:
      regrets = np.sum(
          nu_weights.reshape(-1, 1)
          * self.compute_regrets(use_true_rewards=use_true_rewards),
          axis=0,
      )
    return np.max(regrets) / len(self._nus)

  def get_nus(self):
    return np.array(self._nus)

  def get_mus(self):
    mus = []
    for nu in self._nus:
      mu = utils.MixedDistribution(self._policy_mus, nu)
      mus.append(mu)
    return mus

  def get_rewards(self):
    return self._rewards

  def get_mus_and_weights(self):
    mus = self.get_mus()
    self.normalize_nu_weights()
    return mus, self._nu_weights

  def compute_optimal_eta(self):
    if self._regret_steps_per_step is not None:
      self._eta = min(
          np.sqrt(np.log(len(self._policies)) / self._regret_steps_per_step),
          0.5,
      )

  def normalize_nu_weights(self):
    self._nu_weights = np.array(self._nu_weights) / np.sum(self._nu_weights)

  def get_normalized_nu_weights(self):
    return np.array(self._nu_weights) / np.sum(self._nu_weights)

  def restart(self):
    self._nu_weights = list(self._nu_weights)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.base."""

from absl.testing import absltest

import numpy as np

from open_spiel.python.voting import base


class BaseTest(absltest.TestCase):

  def test_basic_preference_profile(self):
    # Create a preference profile from preferences:
    #
    #    a > b > c > d
    #    b > d > a > c
    #    a > c > d > b
    #    d > b > c > a
    #
    # Each has a weight of 1 by default. E.g. each corresponds to one voter.
    votes = [
        ["a", "b", "c", "d"],
        ["b", "d", "a", "c"],
        ["a", "c", "d", "b"],
        ["d", "b", "c", "a"]
    ]
    profile = base.PreferenceProfile(votes=votes)
    self.assertLen(profile.votes, 4)
    self.assertEqual(profile.total_weight(), 4)

  def test_basic_preference_profile_weighted(self):
    # Create a weighted preference profile from preferences:
    #
    #    1:  a > b > c
    #    2:  a > c > b
    #    3:  b > a > c
    #
    # Each vote has a weight of 1, 2, and 3 respectively.
    votes = [
        base.WeightedVote(1, ["a", "b", "c"]),
        base.WeightedVote(2, ["a", "c", "b"]),
        base.WeightedVote(3, ["b", "a", "c"])
    ]
    profile = base.PreferenceProfile(votes=votes)
    self.assertLen(profile.votes, 3)
    self.assertEqual(profile.total_weight(), 6)

  def test_preference_profile_incremental_group(self):
    # Create a weighted preference profile from preferences:
    #
    #    1:  a > b > c
    #    2:  a > c > b
    #    3:  b > a > c
    #
    # by incrementally adding individual groups and then grouping them.
    profile = base.PreferenceProfile()
    for _ in range(1):
      profile.add_vote(["a", "b", "c"])
    for _ in range(2):
      profile.add_vote(["a", "c", "b"])
    for _ in range(3):
      profile.add_vote(["b", "a", "c"])

    # Assure there are 6 votes, each with weight 1.
    with self.subTest("All votes added correctly"):
      self.assertLen(profile.votes, 6)
      self.assertEqual(profile.total_weight(), 6)
    with self.subTest("Vote weight defaults to 1"):
      for vote in profile.votes:
        self.assertEqual(vote.weight, 1)

    # Group up the votes. Check that there are 3 but with total weight
    # unchanged (6).
    profile.group()
    with self.subTest("Grouping votes reduced to correct number"):
      self.assertLen(profile.votes, 3)
    with self.subTest("Grouping votes did not change total weight"):
      self.assertEqual(profile.total_weight(), 6)
    with self.subTest("Grouping votes computed weights correctly"):
      self.assertEqual(profile.get_weight(["a", "b", "c"]), 1)
      self.assertEqual(profile.get_weight(["a", "c", "b"]), 2)
      self.assertEqual(profile.get_weight(["b", "a", "c"]), 3)

  def test_pref_margin_matrices_strong_condorcet(self):
    votes = [
        base.WeightedVote(1, ["a", "b", "c"]),
        base.WeightedVote(1, ["a", "c", "b"]),
        base.WeightedVote(2, ["c", "a", "b"]),
        base.WeightedVote(1, ["b", "c", "a"]),
    ]
    profile = base.PreferenceProfile(votes=votes)

    pref_matrix = profile.pref_matrix()
    expected_pref_matrix = np.array(
        [[0, 4, 2],
         [1, 0, 2],
         [3, 3, 0]]
    )
    with self.subTest("Preference matrix calculated correctly."):
      self.assertTrue(np.array_equal(pref_matrix, expected_pref_matrix))

    margin_matrix = profile.margin_matrix()
    expected_margin_matrix = np.array(
        [[0, 3, -1],
         [-3, 0, -1],
         [1, 1, 0]]   # <-- all positive, except diagonal:
    )                 #     "c" is a strong Condorcet winner.
    with self.subTest("Expected margin matrix calculated correctly."):
      self.assertTrue(np.array_equal(margin_matrix, expected_margin_matrix))

    # Check that there is exactly one strong Condorcet winner.
    condorcet_winners = profile.condorcet_winner(strong=True,
                                                 margin_matrix=margin_matrix)
    with self.subTest("Exactly one strong Condorcet winner found."):
      self.assertListEqual(condorcet_winners, ["c"])

    # A strong Condorcet winner is also a weak Condorcet winner, by definition.
    condorcet_winners = profile.condorcet_winner(strong=False,
                                                 margin_matrix=margin_matrix)
    with self.subTest("A strong Cond. winner is also a weak Cond. winner."):
      self.assertListEqual(condorcet_winners, ["c"])

  def test_weak_condorcet(self):
    votes = [
        base.WeightedVote(1, ["a", "b", "c"]),
        base.WeightedVote(1, ["a", "c", "b"]),
        base.WeightedVote(1, ["c", "a", "b"]),
        base.WeightedVote(1, ["b", "c", "a"]),
    ]
    profile = base.PreferenceProfile(votes=votes)

    # Leads to margin matrix:
    # [[ 0  2  0]
    #  [-2  0  0]
    #  [ 0  0  0]]
    # ==> no strong Condorcet winners, and two weak Condorcet winners
    margin_matrix = profile.margin_matrix()

    strong_condorcet_winners = profile.condorcet_winner(
        strong=True, margin_matrix=margin_matrix)
    with self.subTest("No strong Condorect winner found."):
      self.assertListEqual(strong_condorcet_winners, [])

    # A strong Condorcet winner is also a weak Condorcet winner, by definition.
    weak_condorcet_winners = profile.condorcet_winner(
        strong=False, margin_matrix=margin_matrix)
    self.assertLen(weak_condorcet_winners, 2)
    with self.subTest("Found all weak Condorcet winners."):
      self.assertCountEqual(["a", "c"], weak_condorcet_winners)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Representation of a distribution for a game.

This is a standard representation for passing distributions into algorithms,
with currently the following implementations:

The main way of using a distribution is to call `value(state)`.
"""

import abc
from typing import Any, Optional

import pyspiel


class Distribution(abc.ABC):
  """Base class for distributions.

  This represents a probability distribution over the states of a game.

  Attributes:
    game: the game for which this distribution is derives
  """

  def __init__(self, game: pyspiel.Game):
    """Initializes a distribution.

    Args:
      game: the game for which this distribution is derives
    """
    self.game = game

  @abc.abstractmethod
  def value(self, state: pyspiel.State) -> float:
    """Returns the probability of the distribution on the state.

    Args:
      state: A `pyspiel.State` object.

    Returns:
      A `float`.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def value_str(self,
                state_str: str,
                default_value: Optional[float] = None) -> float:
    """Returns the probability of the distribution on the state string given.

    Args:
      state_str: A string.
      default_value: If not None, return this value if the state is not in the
        support of the distribution.

    Returns:
      A `float`.
    """
    raise NotImplementedError()

  def __call__(self, state: pyspiel.State) -> float:
    """Turns the distribution into a callable.

    Args:
      state: The current state of the game.

    Returns:
      Float: probability.
    """
    return self.value(state)


class ParametricDistribution(Distribution):
  """A parametric distribution."""

  @abc.abstractmethod
  def get_params(self) -> Any:
    """Returns the distribution parameters."""

  @abc.abstractmethod
  def set_params(self, params: Any):
    """Sets the distribution parameters."""



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Chat bot Arena dataset."""

# pylint: disable=unused-import

import collections
import sys
from absl import app
from absl import flags
import numpy as np
import pandas as pd
import pygraphviz as pgv

from open_spiel.python.utils import gfile

from open_spiel.python.algorithms import nash_averaging
from open_spiel.python.voting import approval
from open_spiel.python.voting import base
from open_spiel.python.voting import borda
from open_spiel.python.voting import copeland
from open_spiel.python.voting import kemeny_young
from open_spiel.python.voting import maximal_lotteries
from open_spiel.python.voting import plurality
from open_spiel.python.voting import ranked_pairs
from open_spiel.python.voting import schulze
from open_spiel.python.voting import stv


SEED = 23875711

# Downloaded from: https://lmsys.org/blog/2023-07-20-dataset/
DATASET_FILE = "/tmp/chatbot_arena_battles.csv"


def parse_battles_dataset(filter_ties=False):
  """Parse the data set from the raw CSV."""
  dataset = []
  model_names = {}
  with gfile.Open(DATASET_FILE, "r") as f:
    lines = f.readlines()
  for line in lines:
    if line.startswith("#"):
      continue
    # ,question_id,model_a,model_b,winner,judge,conversation_a,conversation_b,turn,anony,language,tstamp,openai_moderation,toxic_chat_tag
    parts = line.split(",")
    model_a, model_b, winner = (
        parts[2].strip(),
        parts[3].strip(),
        parts[4].strip(),
    )
    if filter_ties and winner.startswith("tie"):
      continue
    else:
      model_names[model_a] = True
      model_names[model_b] = True
      if winner == "model_a":
        dataset.append((model_a, model_b, -1))
      elif winner == "model_b":
        dataset.append((model_a, model_b, 1))
      else:
        assert winner.startswith("tie")
        dataset.append((model_a, model_b, 0))
  return list(model_names.keys()), dataset


def chatbot_arena_vase(model_names, dataset):
  """Run VasE over Chatbot Arena data set."""

  alternatives = model_names[:]
  profile = base.PreferenceProfile(alternatives=alternatives)
  for datapoint in dataset:
    alt_a, alt_b, outcome = datapoint
    if outcome == 0:
      pass
    elif outcome == -1:
      profile.add_vote([alt_a, alt_b])
    elif outcome == 1:
      profile.add_vote([alt_b, alt_a])

  margin_matrix = profile.margin_matrix()
  strong_cond_winners = profile.condorcet_winner(True, margin_matrix)
  weak_cond_winners = profile.condorcet_winner(False, margin_matrix)
  print(f"Strong Condorcet winner? {strong_cond_winners}")
  print(f"Weak Condorcet winner(s)? {weak_cond_winners}")

  voting_methods = [
      # approval.ApprovalVoting(k=8),
      # borda.BordaVoting(),
      copeland.CopelandVoting(),
      # kemeny_young.KemenyYoungVoting(),
      # Use verbose=True to get more information about the levels
      maximal_lotteries.MaximalLotteriesVoting(iterative=True),
      # maximal_lotteries.MaximalLotteriesVoting(iterative=True, verbose=True),
      # plurality.PluralityVoting(),
      ranked_pairs.RankedPairsVoting(),
      # stv.STVVoting(num_winners=8)
      schulze.SchulzeVoting(),
  ]
  for method in voting_methods:
    print("")
    print(method.name())
    outcome = method.run_election(profile)
    print(outcome.pretty_table_string())
    # print(outcome.pretty_latex_table(header=method.name()))


def ranked_pairs_viz(model_names, dataset):
  """Produce the ranked pairs visualization."""

  alternatives = model_names[:]
  profile = base.PreferenceProfile(alternatives=alternatives)
  num_alternatives = len(alternatives)
  alt_dict = profile.alternatives_dict
  for datapoint in dataset:
    alt_a, alt_b, outcome = datapoint
    if outcome == 0:
      pass
    elif outcome == -1:
      profile.add_vote([alt_a, alt_b])
    elif outcome == 1:
      profile.add_vote([alt_b, alt_a])
  margin_matrix = profile.margin_matrix()
  method = ranked_pairs.RankedPairsVoting()
  outcome = method.run_election(profile)
  graph_mat = outcome.graph
  # Visualize only over the top 8:
  keep_alternatives = [
      "gpt-4",
      "claude-v1",
      "claude-instant-v1",
      "guanaco-33b",
      "gpt-3.5-turbo",
      "wizardlm-13b",
      "palm-2",
      "vicuna-13b",
  ]
  keep_alternatives.sort()
  for j in range(num_alternatives):
    idx = num_alternatives - j - 1
    alt = alternatives[idx]
    if alt not in keep_alternatives:
      graph_mat = np.delete(graph_mat, (idx), axis=0)
      graph_mat = np.delete(graph_mat, (idx), axis=1)
  orig_alternatives = model_names[:]
  alternatives = keep_alternatives
  m = len(alternatives)
  graph = pgv.AGraph(directed=True, strict=True)
  for alternative in alternatives:
    graph.add_node(alternative)
  for i in range(m):
    for j in range(m):
      if graph_mat[i, j] == 1:
        graph.add_edge(alternatives[i], alternatives[j])
        idx_i = alt_dict[alternatives[i]]
        idx_j = alt_dict[alternatives[j]]
        edge = graph.get_edge(
            orig_alternatives[idx_i], orig_alternatives[idx_j]
        )
        edge.attr["label"] = margin_matrix[idx_i, idx_j]
  graph.write("/tmp/chatbot_arena_rps.dot")  # write to simple.dot
  graph.draw(
      "/tmp/chatbot_arena_rps.png",
      # args='-Gdpi=100',
      prog="dot",
  )  # , args="-n2")  # draw
  print("Wrote to /tmp/chatbot_arena_rps.png")


def main(_):
  model_names, dataset = parse_battles_dataset()
  model_names.sort()
  print(f"{len(model_names)} models.")
  print(f"{len(dataset)} datapoints.")
  chatbot_arena_vase(model_names, dataset)
  ranked_pairs_viz(model_names, dataset)


if __name__ == "__main__":
  np.random.seed(SEED)
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests several sequence form utilities."""

from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import cfr
from open_spiel.python.algorithms import expected_game_score as egs
from open_spiel.python.algorithms import sequence_form_utils
import pyspiel

_KUHN_GAME = pyspiel.load_game('kuhn_poker')
_LEDUC_GAME = pyspiel.load_game('leduc_poker')


class SequenceFormTest(parameterized.TestCase):

  @parameterized.parameters(
      {
          'game': _KUHN_GAME,
          'cfr_iters': 100
      },
      {
          'game': _LEDUC_GAME,
          'cfr_iters': 10
      },
  )
  def test_sequence_to_policy(self, game, cfr_iters):

    cfr_solver = cfr.CFRSolver(game)

    for _ in range(cfr_iters):
      cfr_solver.evaluate_and_update_policy()

    (_, infoset_actions_to_seq, infoset_action_maps, _, _,
     _) = sequence_form_utils.construct_vars(game)

    policies = cfr_solver.average_policy()
    sequences = sequence_form_utils.policy_to_sequence(game, policies,
                                                       infoset_actions_to_seq)
    converted_policies = sequence_form_utils.sequence_to_policy(
        sequences, game, infoset_actions_to_seq, infoset_action_maps)
    np.testing.assert_allclose(
        policies.action_probability_array,
        converted_policies.action_probability_array,
        rtol=1e-10)

  @parameterized.parameters(
      {
          'game': _KUHN_GAME,
          'cfr_iters': 100
      },
      {
          'game': _LEDUC_GAME,
          'cfr_iters': 10
      },
  )
  def test_sequence_payoff(self, game, cfr_iters):
    (_, infoset_actions_to_seq, _, _, payoff_mat,
     _) = sequence_form_utils.construct_vars(game)

    uniform_policies = policy.TabularPolicy(game)
    uniform_value = egs.policy_value(
        game.new_initial_state(), [uniform_policies, uniform_policies]
    )
    sequences = sequence_form_utils.policy_to_sequence(game, uniform_policies,
                                                       infoset_actions_to_seq)
    min_mat = -payoff_mat[0]
    np.testing.assert_allclose(
        uniform_value[0],
        -sequences[0].T @ min_mat @ sequences[1],
        rtol=1e-10)

    # use cfr iterations to construct new policy
    cfr_solver = cfr.CFRSolver(game)
    for _ in range(cfr_iters):
      cfr_solver.evaluate_and_update_policy()

    policies = cfr_solver.average_policy()
    cfr_value = egs.policy_value(game.new_initial_state(), [policies, policies])
    sequences = sequence_form_utils.policy_to_sequence(game, policies,
                                                       infoset_actions_to_seq)
    np.testing.assert_allclose(
        cfr_value[0], -sequences[0].T @ min_mat @ sequences[1], rtol=1e-10)

  @parameterized.parameters(
      {
          'game': _KUHN_GAME,
          'seed': 12345
      },
      {
          'game': _LEDUC_GAME,
          'seed': 12345
      },
  )
  def test_sequence_tangent_projection(self, game, seed, step_size=1e-2):
    (_, infoset_actions_to_seq, infoset_action_maps, infoset_parent_map, _,
     _) = sequence_form_utils.construct_vars(game)

    uniform_policies = policy.TabularPolicy(game)
    sequences = sequence_form_utils.policy_to_sequence(game, uniform_policies,
                                                       infoset_actions_to_seq)

    constraints = sequence_form_utils.construct_constraint_vars(
        infoset_parent_map, infoset_actions_to_seq, infoset_action_maps)

    rnd = np.random.RandomState(seed)

    con_errs = []
    for p, (con_mat, b) in constraints.items():
      seq_p = sequences[p]
      # generate random sequence-form "gradient" direction
      grad = rnd.randn(len(seq_p))
      grad = grad / np.linalg.norm(grad)
      # construct tangent projection from sequence-form constraints
      pinv_con_mat = np.linalg.pinv(con_mat)
      proj = np.eye(con_mat.shape[1]) - pinv_con_mat.dot(con_mat)
      # project gradient onto tangent space
      grad_proj = proj.dot(grad)
      # take gradient step (should remain on treeplex w/ small enough step_size)
      seq_p_new = seq_p + step_size * grad_proj
      # measure constraint violation error after gradient step (should be ~0)
      con_err = np.linalg.norm(con_mat.dot(seq_p_new) - b)
      con_errs.append(con_err)

    np.testing.assert_allclose(
        con_errs,
        np.zeros(len(con_errs)),
        atol=1e-10)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A few prompts for extracting the value of a fruit trade with llms.
"""

from open_spiel.python.games.chat_games.envs.payoffs import utils as payoff_utils
from open_spiel.python.games.chat_games.envs.utils import text

MIN_PAYOFF = -20
MAX_PAYOFF = 20
PAYOFF_PROMPT_a = [
    'You are an assistant designed to calculate the values of trades ' +
    'in a fruit trading game. Determine the value of the fruits the player ' +
    'is receiving in the trade. Then determine the value of the fruits the ' +
    'player is giving up through the trade. Subtract the value the player ' +
    'gives away from the value the player receives. Here is an example ' +
    'enclosed by "&".']

PAYOFF_PROMPT_b = [
    '&' * 50,
    'To calculate the trade value, we first calculate the value of ' +
    'the fruit Bob receives in the trade. Bob receives 3 kiwis worth 3 each. ' +
    'Therefore Bob receives a value of 9 in the trade.',
    'Receives: 9',
    'Now we calculate the value of the fruits Bob gives up in the trade. ' +
    'Bob gives up1 banana which is worth 5, therefore, Bob gives up a value ' +
    'of 5 in the trade.',
    'Gives: 5',
    'Subtracting the value Bob gives away from the value Bob receives, we ' +
    'find 9 - 5 = 4.',
    'Calculation: Receives - Gives = 9 - 5 = 4.',
    'Value for Bob: 4.',
    '&' * 50,
    'Now calculate the value of the trade made in the following message.',
    '{m}',
    '&' * 50,
    'Trade calculation for {p} ONLY: ']

PAYOFF_PROMPT = ('\n\n'.join(text.wrap(PAYOFF_PROMPT_a)) + '\n\n' + '&' * 50 +
                 '\n\nBob offered to give up 1 banana for 3 kiwis. Alice ' +
                 'agreed to the trade.\n\n' +
                 '\n\n'.join(text.wrap(PAYOFF_PROMPT_b)))

PAYOFF_OBS_TRANS_PREFIX = ['Read the following dialogue between two parties ' +
                           'attempting to reach a trade agreement. If the ' +
                           'dialogue ends with someone asking a question or ' +
                           'making a couterproposal, an agreement has not ' +
                           'been reached. If the dialogue ends with someone ' +
                           'saying they accept the trade, an agreement has ' +
                           'been reached. Report how much of each fruit each ' +
                           'player gave and received in the tradeby stating ' +
                           'the players names followed by a list of the ' +
                           'fruits the gave up and then a list of the fruits ' +
                           'they received in this format:',
                           'Player [Name]: Receives x Gives y',
                           'Player [Name]: Receives y Gives x',
                           'Example 1:',
                           'Dialogue:',
                           'Bob offered to give up 1 banana for 3 kiwis. ' +
                           'Alice agreed to the trade.',
                           'Player Bob: Receives 3 kiwis Gives 1 banana',
                           'Player Suzy: Receives 1 banana Gives 3 kiwis',
                           'Example 2:',
                           'Dialogue:',
                           'Alice offered to give up 1 banana for 3 kiwis. ' +
                           'George does not want to trade.',
                           'Player Bob: Receives 0 kiwi Gives 0 banana',
                           'Player Suzy: Receives 0 banana Gives 0 kiwi',
                           'Dialogue:']
PAYOFF_OBS_TRANS_PREFIX = ('\n\n'.join(text.wrap(PAYOFF_OBS_TRANS_PREFIX)) +
                           '\n\n')

PAYOFF_OBS_TRANS_POSTFIX = ''

PAYOFF = payoff_utils.Payoff(PAYOFF_PROMPT,
                             MIN_PAYOFF,
                             MAX_PAYOFF,
                             PAYOFF_OBS_TRANS_PREFIX,
                             PAYOFF_OBS_TRANS_POSTFIX)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple basic example."""

# pylint: disable=unused-import

import sys
from absl import app
from absl import flags
import numpy as np

from open_spiel.python.voting import base
from open_spiel.python.voting import copeland


def main(_):
  # Create a preference profile that represents the following votes:
  #   A > B > C
  #   A > C > B
  #   C > A > B
  #   C > A > B
  #   B > C > A
  # This profile has three alternatives: A, B, and C. The strings here "A", "B",
  # "C" represent the alternative's ID and is of type base.AlternativeId.
  # (They can be strings or integers.)
  alternatives = ["A", "B", "C"]

  # Easiest way to make this profile:
  _ = base.PreferenceProfile(alternatives=alternatives, votes=[
      ["A", "B", "C"], ["A", "C", "B"], ["C", "A", "B"], ["C", "A", "B"],
      ["B", "C", "A"]
  ])

  # Note that the C > A > B vote is there twice, so another common way to show
  # this is:
  #   1: A > B > C
  #   1: A > C > B
  #   2: C > A > B
  #   1: B > C > A
  # and can be created with the WeightedVote type directly.
  profile = base.PreferenceProfile(alternatives=alternatives, votes=[
      base.WeightedVote(1, ["A", "B", "C"]),
      base.WeightedVote(1, ["A", "C", "B"]),
      base.WeightedVote(2, ["C", "A", "B"]),
      base.WeightedVote(1, ["B", "C", "A"])
  ])

  # Print some information about the profile
  print(f"Number of alternatives: {profile.num_alternatives()}")
  print(f"Number of votes: {profile.num_votes()}")
  print(f"Alternatives: {profile.alternatives}")
  print("Profile:")
  print(profile)

  # Print a reverse mapping of AlternativeId -> index
  # indices will always be numbered 0 to num_alternatives - 1.
  # Some methods work directly with the indices.
  alt_idx = profile.alternatives_dict
  print("Alternative ids -> index map:")
  print(alt_idx)

  # Iterating through a profile
  print("Iterating through profile:")
  for vote in profile.votes:
    # Each item is a weighted vote:
    print(f"  {vote.weight}: {vote.vote}")

  # Margin matrix and Condorcet winner check
  margin_matrix = profile.margin_matrix()
  cond_winners = profile.condorcet_winner(strong=True,
                                          margin_matrix=margin_matrix)
  print("Margin matrix:")
  print(margin_matrix)
  print(f"Condorcet winners: {cond_winners}")

  # Run Copeland on this profile and print the results
  method = copeland.CopelandVoting()
  outcome = method.run_election(profile)
  print("Copeland outcome:")
  print(outcome.pretty_table_string())


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for crazy eights."""


from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

import pyspiel

# crazy_eights = pyspiel.crazy_eights

NUM_SIM_GAMES = 10
SEED = 87375711


class GamesCrazyEightsTest(parameterized.TestCase):

  def test_crazy_eights_game_sim(self):
    game = pyspiel.load_game("crazy_eights")
    for _ in range(NUM_SIM_GAMES):
      print("----------------")
      print("New game")
      print("----------------")
      state = game.new_initial_state()
      while not state.is_terminal():
        if state.is_chance_node():
          outcomes = state.chance_outcomes()
          action_list, prob_list = zip(*outcomes)
          action = np.random.choice(action_list, p=prob_list)
          print("Chance samples", state.action_to_string(action))
        else:
          print("Player turn")
          print(
              "Dealer's deck:",
              state.get_dealer_deck(),
          )
          actions = state.legal_actions()
          action = np.random.choice(actions)
          print("Action chosen:", state.action_to_string(action))
        state.apply_action(action)
        print("")
      print("Terminal state: ")
      print(str(state))
      print("Returns:", state.returns())
      print("")


if __name__ == "__main__":
  np.random.seed(SEED)
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for euchre."""


from absl.testing import absltest

import pyspiel
euchre = pyspiel.euchre


class GamesEuchreTest(absltest.TestCase):

  def test_bindings(self):
    self.assertEqual(euchre.JACK_RANK, 2)
    self.assertEqual(euchre.NUM_SUITS, 4)
    self.assertEqual(euchre.NUM_CARDS_PER_SUIT, 6)
    self.assertEqual(euchre.NUM_CARDS, 24)
    self.assertEqual(euchre.PASS_ACTION, 24)
    self.assertEqual(euchre.CLUBS_TRUMP_ACTION, 25)
    self.assertEqual(euchre.DIAMONDS_TRUMP_ACTION, 26)
    self.assertEqual(euchre.HEARTS_TRUMP_ACTION, 27)
    self.assertEqual(euchre.SPADES_TRUMP_ACTION, 28)
    self.assertEqual(euchre.GO_ALONE_ACTION, 29)
    self.assertEqual(euchre.PLAY_WITH_PARTNER_ACTION, 30)
    self.assertEqual(euchre.MAX_BIDS, 8)
    self.assertEqual(euchre.NUM_TRICKS, 5)
    self.assertEqual(euchre.FULL_HAND_SIZE, 5)
    game = pyspiel.load_game('euchre')
    state = game.new_initial_state()
    self.assertEqual(state.num_cards_dealt(), 0)
    self.assertEqual(state.num_cards_played(), 0)
    self.assertEqual(state.num_passes(), 0)
    self.assertEqual(state.upcard(), pyspiel.INVALID_ACTION)
    self.assertEqual(state.discard(), pyspiel.INVALID_ACTION)
    self.assertEqual(state.trump_suit(), pyspiel.INVALID_ACTION)
    self.assertEqual(state.left_bower(), pyspiel.INVALID_ACTION)
    self.assertEqual(state.right_bower(), pyspiel.INVALID_ACTION)
    self.assertEqual(state.declarer(), pyspiel.PlayerId.INVALID)
    self.assertEqual(state.declarer_partner(), pyspiel.PlayerId.INVALID)
    self.assertEqual(state.first_defender(), pyspiel.PlayerId.INVALID)
    self.assertEqual(state.second_defender(), pyspiel.PlayerId.INVALID)
    self.assertIsNone(state.declarer_go_alone())
    self.assertEqual(state.lone_defender(), pyspiel.PlayerId.INVALID)
    self.assertEqual(state.active_players(), [True, True, True, True])
    self.assertEqual(state.dealer(), pyspiel.INVALID_ACTION)
    self.assertEqual(state.current_phase(), euchre.Phase.DEALER_SELECTION)
    self.assertEqual(state.current_trick_index(), 0)
    self.assertEqual(state.card_holder(), [None] * 24)
    self.assertEqual(euchre.card_rank(8), euchre.JACK_RANK)
    self.assertEqual(euchre.card_rank(8, euchre.Suit.CLUBS), 100)
    self.assertEqual(euchre.card_suit(8), euchre.Suit.CLUBS)
    self.assertEqual(euchre.card_suit(8, euchre.Suit.SPADES),
                     euchre.Suit.SPADES)
    self.assertEqual(euchre.card_string(8), 'CJ')
    trick = state.tricks()[state.current_trick_index()]
    self.assertEqual(trick.winning_card(), pyspiel.INVALID_ACTION)
    self.assertEqual(trick.led_suit(), euchre.Suit.INVALID_SUIT)
    self.assertEqual(trick.trump_suit(), euchre.Suit.INVALID_SUIT)
    self.assertFalse(trick.trump_played())
    self.assertEqual(trick.leader(), pyspiel.PlayerId.INVALID)
    self.assertEqual(trick.winner(), pyspiel.PlayerId.INVALID)
    self.assertEqual(trick.cards(), [pyspiel.INVALID_ACTION])
    trick = state.current_trick()
    self.assertEqual(trick.led_suit(), euchre.Suit.INVALID_SUIT)
    self.assertEqual(trick.trump_suit(), euchre.Suit.INVALID_SUIT)
    self.assertFalse(trick.trump_played())
    self.assertEqual(trick.leader(), pyspiel.PlayerId.INVALID)
    self.assertEqual(trick.winner(), pyspiel.PlayerId.INVALID)
    self.assertEqual(trick.cards(), [pyspiel.INVALID_ACTION])


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base utils for constructing agent dialogue message headers.
"""

import dataclasses
import string

from typing import Callable, Tuple


@dataclasses.dataclass(frozen=True)
class BaseScenario:
  msg: str
  sender: str
  receiver: str


@dataclasses.dataclass(frozen=True)
class Header:
  plain: str
  w_opts: str
  strip_msg: Callable[[str, str], str]
  special_chars: Tuple[str, ...]
  action_keys: Tuple[str, ...] = tuple([])
  info_keys: Tuple[str, ...] = tuple([])
  context: str = ''


def plain_header_is_valid(header: Header) -> bool:
  plain = header.plain
  keys = [t[1] for t in string.Formatter().parse(plain) if t[1] is not None]
  return 'sender' in keys and 'receiver' in keys



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exploitability measurement utils for symmetric games."""

from absl import logging  # pylint:disable=unused-import

import numpy as np
from scipy import special

from open_spiel.python.algorithms.adidas_utils.helpers import misc


def unreg_exploitability(dist, payoff_tensor):
  """Compute exploitability of dist for symmetric game.

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    payoff_tensor: (>=1 x A x ... x A) np.array, payoffs for each joint action
  Returns:
    exploitability (float): payoff of best response - payoff of dist
  """
  num_players = payoff_tensor.shape[0]
  nabla = misc.pt_reduce(payoff_tensor[0], [dist] * num_players, [0])

  u_br = np.max(nabla)
  u_dist = nabla.dot(dist)

  return u_br - u_dist


def ate_exploitability(dist, payoff_tensor, p=1):
  """Compute Tsallis regularized exploitability of dist for symmetric game.

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    payoff_tensor: (>=1 x A x ... x A) np.array, payoffs for each joint action
      assumed to be non-negative
    p: float in [0, 1], Tsallis entropy-regularization --> 0 as p --> 0
  Returns:
    exploitability (float): payoff of best response - payoff of dist
  """
  if payoff_tensor.min() < 0.:
    raise ValueError('payoff tensor must be non-negative')
  num_players = payoff_tensor.shape[0]
  nabla = misc.pt_reduce(payoff_tensor[0], [dist] * num_players, [0])
  if p > 0:
    power = 1./p
    s = np.linalg.norm(nabla, ord=power)
    br = (nabla / np.linalg.norm(nabla, ord=power))**power
  else:
    power = np.inf
    s = np.linalg.norm(nabla, ord=power)
    br = np.zeros_like(dist)
    maxima = (nabla == s)
    br[maxima] = 1. / maxima.sum()

  u_br = nabla.dot(br) + s / (p + 1) * (1 - np.sum(br**(p + 1)))
  u_dist = nabla.dot(dist) + s / (p + 1) * (1 - np.sum(dist**(p + 1)))

  return u_br - u_dist


def qre_exploitability(dist, payoff_tensor, temperature=0.):
  """Compute Shannon regularized exploitability of dist for symmetric game.

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    payoff_tensor: (>=1 x A x ... x A) np.array, payoffs for each joint action
      assumed to be non-negative
    temperature: non-negative float
  Returns:
    exploitability (float): payoff of best response - payoff of dist
  """
  num_players = payoff_tensor.shape[0]
  nabla = misc.pt_reduce(payoff_tensor[0], [dist] * num_players, [0])
  if temperature > 0:
    br = special.softmax(nabla / temperature)
  else:
    br = np.zeros_like(dist)
    maxima = (nabla == np.max(nabla))
    br[maxima] = 1. / maxima.sum()

  u_br = nabla.dot(br) + temperature * special.entr(br).sum()
  u_dist = nabla.dot(dist) + temperature * special.entr(dist).sum()

  return u_br - u_dist


def grad_norm_exploitability(dist, payoff_tensor, eta=None, temperature=0.):
  """Compute (avg, max) exploitability of dist for non-symmetric game.

  Args:
    dist: 1-d np.array, current estimate of nash distribution
    payoff_tensor: (>=1 x A x ... x A) np.array, payoffs for each joint action
      assumed to be non-negative
    eta: step size for approximate best response (default 1 / (n * m))
      where n is # of players and m is # of actions (same for all players)
    temperature: non-negative float
  Returns:
    exploitability (float): squared norm of projected-gradient
  """

  if eta is None:
    eta = 1. / dist.size

  num_players = payoff_tensor.shape[0]
  nabla = misc.pt_reduce(payoff_tensor[0], [dist] * num_players, [0])
  if temperature > 0:
    nabla -= temperature * (np.log(dist) + 1)

  m = dist.size
  nabla_proj = nabla - 1. / m * np.sum(nabla)
  nabla_sq_norm = np.inner(nabla_proj, nabla_proj)

  return eta * nabla_sq_norm
